from typing import List, Type
import logging

from ..ACMEobj import ACMEOrder, ACMEAuthorization
from ..util import get_dns_chall_txt_record
from .dnsproviderbase import _DNSProviderBase
from . import aliyun


logger = logging.getLogger(__name__)
info = logger.info
debug = logger.debug


class DNS01ChallengeRespondHandler:
    """
    use `eval(dnsprovider.Handler)` to get a dns handler class; to add more
    dns providers, name the new provider module using name that will be added 
    to attribute `supported_provider`, and create `Handler` class by 
    inheriting `dnsproviderbase._DNSProviderBase`.
    """

    supported_provider = ['aliyun']

    def __init__(self, 
                 order_obj: ACMEOrder,
                 dnsprovider: str,
                 access_key: str,
                 secret: str,
                 **kwargs):
        if not dnsprovider in self.supported_provider:
            raise ValueError(f'dnsprovider {dnsprovider} not supported')
        self.order_obj = order_obj
        handler_cls: Type[_DNSProviderBase] = eval(f'{dnsprovider}.Handler')
        self.handler = handler_cls(access_key, secret, **kwargs)
    
    def dns_chall_respond(self) -> List[ACMEAuthorization]:
        """add dns-01 respond to selected dns provider"""
        for auth in self.order_obj.auth_objs:
            if auth.chall_dns.status == 'valid':
                info(f'challenge for {auth.identifier_value} is already valid')
                continue
            value = get_dns_chall_txt_record(
                token=auth.chall_dns.token,
                jwk=self.order_obj.related_acct.jwk_obj
            )
            self.handler.add_txt_record(auth.identifier_value, value)
            auth.chall_dns.respond()
            info(f'respond to dns challenge for {auth.identifier_value}')
        return self.order_obj.auth_objs
    
    def clear_dns_record(self) -> None:
        self.handler.clear_dns_record()



import base64
import logging
import socketserver
import hashlib
import json
import sys
import time
from argparse import Namespace
from http.server import SimpleHTTPRequestHandler
from typing import Any, Dict, List, Union
from pathlib import Path
from zipfile import ZipFile

from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.x509 import NameAttribute, DNSName, SubjectAlternativeName
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa, ec
import requests

from pyacme.base import _JWKBase
from pyacme.jwk import JWKRSA, JWKES256
from pyacme.settings import *


# only handle log record with level DEBUG
debug_hd = logging.StreamHandler(stream=sys.stdout)
debug_hd.setFormatter(logging.Formatter(LOG_DEBUG_FMT, style='{'))
debug_hd.setLevel(logging.DEBUG)
debug_hd.addFilter(lambda r: r.levelno == logging.DEBUG)

# handle log level INFO and above
info_hd = logging.StreamHandler(stream=sys.stdout)
info_hd.setFormatter(logging.Formatter(LOG_FMT, style='{'))
info_hd.setLevel(LOG_LEVEL)

base_logger = logging.getLogger('pyacme')
base_logger.addHandler(debug_hd)
base_logger.addHandler(info_hd)
base_logger.setLevel(LOG_LEVEL)

logger = logging.getLogger(__name__)
info = logger.info


# test logging
# def info_test(msg): info(msg)
# def debug_test(msg): logger.debug(msg)


def get_keyAuthorization(token: str, jwk: _JWKBase) -> str:
    """
    construct auth string by joining challenge token and key thumbprint.

    see https://tools.ietf.org/html/rfc8555#section-8.1
    """
    # see https://github.com/diafygi/acme-tiny/blob/master/acme_tiny.py#L86
    # sort keys required by https://tools.ietf.org/html/rfc7638#section-4
    s_jwk = json.dumps(jwk._container, sort_keys=True, separators=(',', ':'))
    jwk_hash = hashlib.sha256(s_jwk.encode(encoding='utf-8')).digest()
    b64 = base64.urlsafe_b64encode(jwk_hash).strip(b'=')
    return f"{token}.{str(b64, encoding='utf-8')}"


def get_dns_chall_txt_record(token: str, jwk: _JWKBase) -> str:
    """
    return a string of a dns TXT record; the whole dns record looks like
    `_acme-challenge.www.example.org. 300 IN TXT keyauth_digest`

    see https://tools.ietf.org/html/rfc8555#section-8.4
    """
    keyauth = get_keyAuthorization(token, jwk)
    # sha256 on the keyauth string, see rfc8555 8.4 p66
    keyauth_digest = hashlib.sha256(keyauth.encode('utf-8')).digest()
    b64 = base64.urlsafe_b64encode(keyauth_digest).strip(b'=')
    return str(b64, encoding='utf-8')


def generate_rsa_privkey(privkey_dir: str, 
                         keysize = 2048,
                         key_name = 'certkey.key') -> rsa.RSAPrivateKey:
    """
    generate private key to specified dir using `cryptography` package
    """
    # create a private key if not given
    csr_priv_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=keysize,
        backend=default_backend()
    )
    csr_priv_key_b = csr_priv_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption()
    )
    with open(f'{privkey_dir}/{key_name}', 'wb') as f:
        f.write(csr_priv_key_b)
    return csr_priv_key


def generate_ecdsa_privkey(privkey_dir: str, 
                           curve = ec.SECP256R1(),
                           key_name = 'certkey.key'
                           ) -> ec.EllipticCurvePrivateKey:
    """
    generate ecdsa key with SECP256R1 to specified dir
    """
    # create a private key if not given
    csr_priv_key = ec.generate_private_key(
        curve=curve,
        backend=default_backend()
    )
    csr_priv_key_b = csr_priv_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption()
    )
    with open(f'{privkey_dir}/{key_name}', 'wb') as f:
        f.write(csr_priv_key_b)
    return csr_priv_key


def generate_privkey(key_type: str,
                     privkey_dir: str, 
                     key_name = 'certkey.key',
                     key_size = 2048
                     ) -> Union[ec.EllipticCurvePrivateKey, rsa.RSAPrivateKey]:
    if key_type not in set(CSR_SUPPORTED_KEY_TYPE+KEY_ACCT_KEYTYPE):
        raise ValueError(f'not supported privated key type {key_type}')
    if key_type == 'rsa':
        return generate_rsa_privkey(privkey_dir, key_size, key_name)
    elif key_type == 'es256':
        return generate_ecdsa_privkey(privkey_dir, ec.SECP256R1, key_name)
    else:
        raise NotImplementedError(f'key_type {key_type} not implemented')
    # other keytype


def create_csr(privkey: Union[rsa.RSAPrivateKey, ec.EllipticCurvePrivateKey],
               domains: List[str],
               *, 
               C = '', 
               ST = '', 
               L = '', 
               O = '', 
               OU = '', 
               emailAddress = '') -> bytes:
    """
    generate csr using `cryptography.x509`;
    """
    csr = x509.CertificateSigningRequestBuilder()
    cn = domains[0]
    csr = csr.subject_name(
        x509.Name(
            [
                NameAttribute(NameOID.COUNTRY_NAME, C),
                NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, ST),
                NameAttribute(NameOID.LOCALITY_NAME, L),
                NameAttribute(NameOID.ORGANIZATION_NAME, O),
                NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, OU),
                NameAttribute(NameOID.COMMON_NAME, cn),
                NameAttribute(NameOID.EMAIL_ADDRESS, emailAddress),
            ]
        )
    )
    alt_names = tuple(DNSName(d) for d in domains)
    csr = csr.add_extension(SubjectAlternativeName(alt_names), critical=False)
    csr_signed = csr.sign(
        privkey, algorithm=hashes.SHA256(), backend=default_backend()
    )
    return csr_signed.public_bytes(serialization.Encoding.DER)


def parse_csr(privkey: Union[rsa.RSAPrivateKey, ec.EllipticCurvePrivateKey], 
              domains: List[str], 
            #   extra: List[str] = [], 
            #   engine: str = 'openssl',
              **subjects: str) -> bytes:
    """
    ouput DER format bytes of a CSR, subjects for csr list below:
     * C = Country two-digit, like GB or US;
     * ST = State or Province
     * L  = Locality
     * O  = Organization Name        
     * OU = Organizational Unit Name
     * emailAddress = test@email.address

    """
    return create_csr(privkey, domains, **subjects)


def save_cert(cert_resp: requests.Response, cert_dir: str) -> requests.Response:
    """
    return 3 cert files 
    as below
     * `cert.pem` the server cert file;
     * `chain.pem` intermediate cert file;
     * `fullchain.pem` both the cert and intermediate, as reponse by the ACME
     server
    """
    fullchain = cert_resp.text
    fullchain_path = Path(cert_dir).absolute() / 'fullchain.pem'
    with open(f'{fullchain_path!s}', 'w') as f:
        f.write(fullchain)
    
    cert, chain = fullchain.split('-----END CERTIFICATE-----\n', maxsplit=1)
    cert += '-----END CERTIFICATE-----\n' 

    cert_path = Path(cert_dir).absolute() / 'cert.pem'
    with open(f'{cert_path!s}', 'w') as f:
        f.write(cert)
    
    chain_path = Path(cert_dir).absolute() / 'chain.pem'
    with open(f'{chain_path!s}', 'w') as f:
        f.write(chain)

    return cert_resp


def run_http_server(path: Union[Path, str], port = 80) -> None:
    """run a pyhton http server on port 80 to reponse acme challenge"""
    class Handler(SimpleHTTPRequestHandler):
        def __init__(self, *args, **kwargs) :
            super().__init__(*args, directory=str(path), **kwargs)

    with socketserver.TCPServer(
        ('', port), Handler, bind_and_activate=False) as httpd:

        # TODO proper exit for http server

        # prevent "OSError: [Errno 98] Address already in use" when testing
        httpd.allow_reuse_address = True
        httpd.server_bind()
        httpd.server_activate()
        info(f'serving at port {port}')
        httpd.serve_forever()


def jwk_factory(acct_priv_key_path: str) -> _JWKBase:
    """generate jwk object according private key file"""
    with open(acct_priv_key_path, 'rb') as f:
        acct_priv = serialization.load_pem_private_key(
            data=f.read(),
            password=None,
            backend=default_backend()
        )
        if isinstance(acct_priv, rsa.RSAPrivateKey):
            jwk = JWKRSA(
                priv_key=acct_priv,
                n=acct_priv.public_key().public_numbers().n,
                e=acct_priv.public_key().public_numbers().e
            )
        elif isinstance(acct_priv, ec.EllipticCurvePrivateKey):
            if isinstance(acct_priv.curve, ec.SECP256R1):
                jwk = JWKES256(acct_priv)
            else:
                raise NotImplementedError(
                    f'ecdsa curve {acct_priv.curve} not implemented'
                )
        else:
            raise TypeError(f'key type {type(acct_priv)} not supported')
        return jwk


def backup_certs(cert_path: str, backup_path: str) -> None:
    if not list(Path(cert_path).iterdir()):
        # print('no backup performed')
        info('no backup performed')
        return
    date_time = time.strftime(BAK_TIME_FMT, time.localtime())
    bak_zip_name = BAK_DEFAULT_PATTERN.format(date_time=date_time)
    with ZipFile(Path(backup_path)/bak_zip_name, 'w') as zip_f:
        for f in Path(cert_path).iterdir():
            zip_f.write(str(f))
    info(f'certificates backup zipped to {Path(backup_path)/bak_zip_name}')


def domains_check(domains: List[str]) -> List[str]:
    """
    all input domains should have the same primary, 
    e.g.
     * `example.com`, `a.example.com` are valid
     * `example.com`, `a.example1.com` are invalid
     * `domains=["a.x.com", "b.x.com"]` will be converted to 
     `["x.com", "a.x.com", "b.x.com"]`
    if top-level like `example.com` is not given in `domains`, it will be 
    added here
    """
    if len(domains) == 1:
        return domains
    primary_set: List[str] = []
    for d in domains:
        primary = '.'.join(d.split('.')[-2:])
        primary_set.append(primary)
    if len(set(primary_set)) == 1:
        primary = primary_set.pop()
        # if top-level given by user but not in the first
        if primary in domains:
            # move primary to the first element in domains list
            domains_mod = sorted(domains, key=len)
            return domains_mod
        else:
            # if top-level is not given by user, add it to domains
            return [primary] + domains
    else:
        raise ValueError(f'input domains {domains} have different top-level')


def check_path(wd: str, domains: List[str]) -> str:
    """
    for default file structure, `root="~/.pyacme"`; may be subsitituted by user
    given value.
    ```
    {root}/{domain_name}
    +-- acct
    |   +-- acct.pem
    +-- chall_http
    |   +-- {http_chall_token}
    |   +-- ...
    +-- cert
    |   +-- cert.key
    |   +-- chain.pem
    |   +-- fullchain.pem
    +-- backup
    |   +-- bak_{date_time}.zip
    |   +-- ...
    ```
    multiple domains will be concatenated and placed under one directory
    """
    d = '_'.join(domains)
    wd_path = Path(wd).expanduser().absolute() / d
    if not wd_path.exists():
        wd_path.mkdir(parents=True, exist_ok=True)
        (wd_path / WD_ACCT).mkdir()
        (wd_path / WD_HTTP01).mkdir()
        (wd_path / WD_CERT).mkdir()
        (wd_path / WD_BAK).mkdir()
        acme_http = Path('.well-known/acme-challenge')
        (wd_path / WD_HTTP01 / acme_http).mkdir(parents=True)
        info(f'workding directory {wd_path!s} created')
    else:
        info(f'workding directory {wd_path!s} exists')
    return d


def create_new_acct_key(key_path: Path, acct_key_type: str) -> None:
    # passed in key_path is a file location
    if not key_path.exists():
        key_path = key_path.parent
        generate_privkey(acct_key_type, str(key_path), KEY_ACCT)
        info(f'new {acct_key_type} account private key generated at {key_path}')
    else:
        info(f'use existed account private key at {key_path}')


def main_param_parser(args: Namespace) -> dict:
    """
    parse params passed to `main()`, assign proper default value if needed;
    perform `check_path()` at the end of this parser;
    arg `--debug` is not included here
    """
    # set logging level
    if args.debug:
        logging.getLogger('pyacme').setLevel(logging.DEBUG)

    domains = domains_check(args.domain)

    # use original given domains for path creation
    joined_domain = check_path(args.working_directory, args.domain)

    param_dict: Dict[str, Any] = dict()

    # wildcard domain only available for dns mode
    param_dict['domains'] = domains
    for d in args.domain:
        if '*' in d:
            param_dict['mode'] = 'dns'
            break
    else:
        param_dict['mode'] = args.mode
    
    # create new acct key if not exist in working dir
    wd = Path(args.working_directory).expanduser().absolute() / joined_domain
    if not args.account_private_key:
        key_path = Path(wd) / WD_ACCT / KEY_ACCT
        create_new_acct_key(key_path, args.acct_key_type)
        param_dict['acct_priv_key'] = str(key_path)
    else:
        param_dict['acct_priv_key'] = args.account_private_key
        info(f'use user given account key at {args.account_private_key}')

    # parse param for cert_path and chall_path
    param_dict['cert_path'] = str(wd / WD_CERT)
    param_dict['chall_path'] = str(wd / WD_HTTP01)
            
    # parse subject names string for CSR, append country code to csr_dict
    csr_dict = {'C': args.country_code}
    if args.csr_subjects:
        csr_list = args.csr_subjects.split(',')
        csr_dict.update({i[0]:i[1] for i in [j.split('=') for j in csr_list]})
    param_dict['subject_names'] = csr_dict

    # parse dns specifics
    dns_dict = dict()
    if args.dns_specifics:
        dns_list = args.dns_specifics.split(',')
        dns_dict = {i[0]:i[1] for i in [j.split('=') for j in dns_list]}
    param_dict['dns_specifics'] = dns_dict

    # direct pass params
    key_list = [
        'contact',
        'not_before',
        'not_after',
        'dnsprovider',
        'access_key',
        'secret',
        'CA_entry',
        'poll_interval',
        'poll_retry_count',
        'csr_priv_key_type',
        'csr_priv_key_size',
        'chall_resp_server_port', 
        'no_ssl_verify'
    ]
    for key in key_list:
        param_dict[key] = getattr(args, key)
    
    # backup old cert files if exist
    backup_certs(str(wd/WD_CERT), str(wd/WD_BAK))

    return param_dict


import time

import pytest

from conftest import rsa, Path
from conftest import ACMERequestActions, JWSRS256, JWKRSA, _JWKBase, \
                     ACMEAccount, ACMEAccountActions, ACMEError, ACMEOrder, \
                     settings
from test_common import *


def count_equal(a_list: list, b_list: list) -> None:
    """assert equality of two lists, order not considered"""
    a_list, b_list = sorted(a_list), sorted(b_list)
    assert len(a_list) == len(b_list)
    assert all([a == b for a, b in zip(a_list, b_list)])


class TestACMERequestActionsTest:

    def test_retry_badNonce(self,
                            new_request_action: ACMERequestActions,
                            new_rsa_privkey: rsa.RSAPrivateKey):
        """
        create account without using ACMEAccountAction, while using incorrect
        nonce to trigger badNonce retry.
        """
        req = new_request_action
        jws = JWSRS256(
            url=req.acme_dir['newAccount'],
            nonce='badNonce',
            jwk = JWKRSA(
                priv_key=new_rsa_privkey,
                n=new_rsa_privkey.public_key().public_numbers().n,
                e=new_rsa_privkey.public_key().public_numbers().e
            ),
            payload={
                'termsOfServiceAgreed': True,
                'contact': TEST_CONTACTS
            },
        )
        jws.sign()
        resp = req.new_account(jws)
        assert resp.status_code == 201


class TestACMEAccountObjectCreation:

    def test_init_by_create(self, 
                            new_jwk: _JWKBase,
                            new_acct_action: ACMEAccountActions):
        acct_obj = ACMEAccount.init_by_create(
            jwk=new_jwk,
            acct_actions=new_acct_action,
            contact=TEST_CONTACTS
        )
        # reponse 201-created if an account is created
        assert acct_obj._resp.status_code == 201

    def test_init_by_query(self, 
                           new_acct_action: ACMEAccountActions,
                           new_acct_obj: ACMEAccount):
        acct_obj_query = ACMEAccount.init_by_query(
            jwk=new_acct_obj.jwk_obj,
            acct_actions=new_acct_action
        )
        # reponse 200-OK if an account is returned successfully
        assert acct_obj_query._resp.status_code == 200


class TestACMEAccountActions:

    def test_poll_acct_state(self, new_acct_obj: ACMEAccount):
        # for an acct just created, status code will be 201
        assert new_acct_obj._resp.status_code == 201
        new_acct_obj.poll_acct_state()
        # poll and update acct state, status code will be updated to 200
        assert new_acct_obj._resp.status_code == 200
    
    def test_update_account(self, new_acct_obj: ACMEAccount):
        new_acct_obj.update_account(contact=TEST_CONTACTS_MOD)
        # successful update will return 200-OK
        assert new_acct_obj._resp.status_code == 200
        new_acct_obj.poll_acct_state()
        # assert if attribute "contact" is actually updated
        count_equal(new_acct_obj.contact, TEST_CONTACTS_MOD)
    
    def test_deactivate(self, new_acct_obj: ACMEAccount):
        new_acct_obj.deactivate()
        # successful deactivation will return 200-OK
        assert new_acct_obj._resp.status_code == 200
        assert new_acct_obj.status == 'deactivated'
        # now post-as-get to a deactivated acct will have 403-Forbidden
        with pytest.raises(ACMEError) as caught:
            new_acct_obj.poll_acct_state()
        assert caught.value.status_code == 403
    
    def test_account_key_rollover(self, 
                                  new_jwk_i: _JWKBase,
                                  new_acct_obj: ACMEAccount):
        new_acct_obj.account_key_rollover(new_jwk_i)
        # successful rollover will return 200-OK
        assert new_acct_obj._resp.status_code == 200
        assert new_acct_obj.jwk_obj is new_jwk_i
        with pytest.raises(ACMEError) as caught:
            new_acct_obj.account_key_rollover(new_jwk_i)
        assert "New and old key are identical" in str(caught.value)


class TestACMEOrder:

    domain = ['i.test.local']
    domains = ['a.test.local', 'b.test.local']

    def _test_new_order_domain(self, d: list, new_acct_obj: ACMEAccount):
        order_obj = new_acct_obj.new_order(identifiers=d)
        # successful order creation will return 201-created
        assert new_acct_obj._resp.status_code == 201
        assert order_obj._resp.status_code == 201
        # only one order related to the account
        assert len(new_acct_obj.order_objs) == 1
        assert new_acct_obj.order_objs[0] is order_obj
        count_equal(d, order_obj.identifier_values)
    
    def test_new_order_single_domain(self, new_acct_obj: ACMEAccount):
        self._test_new_order_domain(self.domain, new_acct_obj)

    def test_new_order_multi_domain(self, new_acct_obj: ACMEAccount):
        self._test_new_order_domain(self.domains, new_acct_obj)
    
    def test_poll_order_state(self, new_order_obj: ACMEOrder):
        new_order_obj.poll_order_state()
        assert new_order_obj._resp.status_code == 200
    
    def test_get_orders(self, 
                        new_acct_action: ACMEAccountActions,
                        new_acct_obj: ACMEAccount,
                        new_order_obj: ACMEOrder):
        assert len(new_acct_obj.order_objs) == 1
        # create new orders directly from account actions
        new_acct_action.new_order(
            acct_obj=new_acct_obj,
            identifiers=[dict(type='dns', value=v) for v in self.domains],
            not_before='',
            not_after='',
            jws_type=new_acct_obj.jwk_obj.related_JWS
        )
        # refresh account order list
        new_acct_obj.get_orders()
        assert len(new_acct_obj.order_objs) == 2


class TestACMEAuthorization:

    domain = ['a.test.local', 'b.test.local']

    def test_poll_auth_state(self, new_order_obj: ACMEOrder):
        assert len(new_order_obj.auth_objs) == 2
        for auth in new_order_obj.auth_objs:
            auth.poll_auth_state()
            assert auth._resp.status_code == 200
    
    def test_deactivate_auth(self, new_order_obj: ACMEOrder):
        auth_to_deactivate = new_order_obj.auth_objs[0]
        auth_to_deactivate.deactivate_auth()
        # successful deactivation will return 200-OK
        assert auth_to_deactivate._resp.status_code == 200
        assert auth_to_deactivate.status == 'deactivated'
        # related order state will also become "deactivated"
        new_order_obj.poll_order_state()
        assert new_order_obj.status == 'deactivated'


class TestACMEChallenge:
    """respond using pebble challtest"""

    # use one order with 1 auth to save test time
    domain = ['a.test.local']

    def test_respond_http(self, new_order_obj: ACMEOrder):
        jwk = new_order_obj.related_acct.jwk_obj
        for auth in new_order_obj.auth_objs:
            add_http_01(auth.chall_http.token, jwk)
            auth.chall_http.respond()
            # if poll auth immediately after repond, status should be "pending"
            auth.poll_auth_state()
            assert auth.status == 'pending'
            time.sleep(4)
            auth.poll_auth_state()
            assert auth.chall_http.status == 'valid'
            assert auth.status == 'valid'
        # once all auth valid, order state become "ready"
        new_order_obj.poll_order_state()
        assert new_order_obj.status == 'ready'

    @staticmethod
    def _test_respond_dns(new_order_obj: ACMEOrder):
        jwk = new_order_obj.related_acct.jwk_obj
        for auth in new_order_obj.auth_objs:
            add_dns_01(auth.chall_dns.token, auth.identifier_value, jwk)
            auth.chall_dns.respond()
            # if poll auth immediately after repond, status should be "pending"
            auth.poll_auth_state()
            assert auth.status == 'pending'
            time.sleep(5)
            auth.poll_auth_state()
            assert auth.chall_dns.status == 'valid'
            assert auth.status == 'valid'
        # once all auth valid, order state become "ready"
        new_order_obj.poll_order_state()
        assert new_order_obj.status == 'ready'
    
    def test_respond_dns(self, new_order_obj: ACMEOrder):
        self._test_respond_dns(new_order_obj)
    
    @pytest.mark.domain('*.test.local')
    def test_respond_dns_wildcard(self, new_order_obj: ACMEOrder):
        print(new_order_obj.identifier_values)
        self._test_respond_dns(new_order_obj)


class TestACMEOrderCertificate:

    domain = ['a.test.local']

    def test_download_cert(self, 
                           new_tmp_wd: Tuple[Path, Path, Path, Path], 
                           new_rsa_privkey_i: rsa.RSAPrivateKey,
                           new_ready_order: ACMEOrder):
        new_ready_order.poll_order_state()
        assert new_ready_order.status == 'ready'
        wd, cert_path, chain_path, fullchain_path = new_tmp_wd
        download_root_cert(wd)
        # test finalize order
        new_ready_order.finalize_order(
            privkey=new_rsa_privkey_i,
            emailAddress='email@address.test',
            C='CN',
            ST='test ST',
            L='test L',
            O='test org',
            OU='test OU'
        )
        time.sleep(1)
        new_ready_order.poll_order_state()
        # after finalized, state become "valid", cannot be finalized again
        assert new_ready_order.status == 'valid'

        # test download cert
        download_resp = new_ready_order.download_certificate(str(wd))
        assert download_resp.status_code == 200

        # external openssl verify
        completed_p = openssl_verify(cert_path, chain_path, wd)
        assert completed_p.returncode == 0


"""see https://tools.ietf.org/html/rfc8555#section-6.7"""

from typing import List, Dict, Any
import json

import requests

# from pyacme.ACMEobj import Empty


class ACMEError(Exception):
    """see https://tools.ietf.org/html/rfc8555#section-6.7"""

    def __init__(self, resp: requests.Response) -> None:
        self.status_code: int
        self.type = ''
        self._resp = resp
        self.subproblems: List[Dict[str, Any]] = []
        self._update_exeception(resp)

    def _update_exeception(self, resp: requests.Response) -> None:
        cls = type(self)
        self._status_code = resp.status_code
        if resp.text:
            _content = json.loads(resp.text)
            # type e.g. "urn:ietf:params:acme:error:badCSR"
            if 'type' in _content:
                self.type = _content['type'].split(':')[-1]
                _content.pop('type')
            if 'subproblems' in _content:
                self.subproblems = _content['subproblems']
                _content.pop('subproblems')
            # load rest attributes
            self.__dict__.update(_content)
            # if badNonce error raised, new nonce should be returned by server
            if 'Replay-Nonce' in resp.headers:
                self.new_nonce = resp.headers['Replay-Nonce']
    
    @property
    def status_code(self) -> int:
        return self._status_code
    
    def __str__(self) -> str:
        cls = type(self).__name__
        _dict = {
            k : v for (k, v) in self.__dict__.items() if not k.startswith('_')
        }
        s = f'{cls}[{self.type} {self._status_code}]({str(_dict)})'
        return s


from pathlib import Path


# run a test pebble server in docker
# https://github.com/letsencrypt/pebble

# to prevent client ssl error, append test/certs/pebble.minica.pem to python
# package `certifi` cacert.pem which is used by `requests` or
# set "verify=False" when using requests
# see https://github.com/letsencrypt/pebble/tree/master/test/certs
VERIFY_SSL = False

TEST_IP = "127.0.0.1"
TEST_PORT = 14000

PEBBLE_TEST = f"https://{TEST_IP}:{TEST_PORT}/dir"
 
# use pebble verify port
PY_HTTPSERVER_PORT = '5002'

# pebble challenge test server management interfaces
# see https://github.com/letsencrypt/pebble/tree/master/cmd/pebble-challtestsrv
CHALL_TEST_PORT = 8055

PEBBLE_CHALLTEST_DNS_A = f"http://{TEST_IP}:{CHALL_TEST_PORT}/add-a"
PEBBLE_CHALLTEST_DNS_A_DEL = f"http://{TEST_IP}:{CHALL_TEST_PORT}/clear-a"

PEBBLE_CHALLTEST_HTTP01 = f"http://{TEST_IP}:{CHALL_TEST_PORT}/add-http01"
PEBBLE_CHALLTEST_HTTP01_DEL = f"http://{TEST_IP}:{CHALL_TEST_PORT}/del-http01"

PEBBLE_CHALLTEST_DNS01 = f"http://{TEST_IP}:{CHALL_TEST_PORT}/set-txt"
PEBBLE_CHALLTEST_DNS01_DEL = f"http://{TEST_IP}:{CHALL_TEST_PORT}/clear-txt"

PEBBLE_CHALLTEST_TLS01 = f"http://{TEST_IP}:{CHALL_TEST_PORT}/add-tlsalpn01"
PEBBLE_CHALLTEST_TLS01_DEL = f"http://{TEST_IP}:{CHALL_TEST_PORT}/del-tlsalpn01"

# name for pebble and challtest container
PEBBLE_DOCKER_FILE = \
    Path(__file__).parents[2].absolute() / 'pebble' / 'docker-compose.yml'
PEBBLE_CONTAINER = 'pebble_pebble_1'
PEBBLE_CHALLTEST_CONTAINER = 'pebble_challtestsrv_1'

# staging root ca
STAGING_ROOT_CA = 'https://letsencrypt.org/certs/fakelerootx1.pem'

## test
# print(Path(__file__).parents[2])
# print(__file__)
# print(PEBBLE_DOCKER_FILE)


# url to acme server's /directory
LETSENCRYPT_PRODUCTION = "https://acme-v02.api.letsencrypt.org/directory"
LETSENCRYPT_STAGING = "https://acme-staging-v02.api.letsencrypt.org/directory"

# working directory settings
WD_DEFAULT = '~/.pyacme'
WD_HTTP01 = 'chall_http'
WD_CERT = 'cert'
WD_ACCT = 'acct'
WD_BAK = 'backup'

# keys related
KEY_ACCT_KEYTYPE = ['rsa', 'es256']
KEY_ACCT_KEYTYPE_DEFAULT = 'rsa'
KEY_ACCT = 'acct.pem'
KEY_SIZE = 2048

# certificate related
CERT_NAME = 'cert.pem'
CERT_CHAIN = 'chain.pem'
CERT_FULLCHAIN = 'fullchain.pem'

# challenge related
MODES_SUPPORTED = ['http', 'dns']
MODES_DEFAULT = 'dns'

# HTTP01
HTTP_SERVING_PORT = 80

# DNS01 and dns provider
DNS_PROVIDERS = ['aliyun']
DNS_DEFAULT_PROVIDER = 'aliyun'

# CSR related
CSR_SUPPORTED_KEY_TYPE = ['rsa', 'es256']
CSR_DEFAULT_KEY_TYPE = 'rsa'
CSR_RSA_PRIVATE_KEY_SIZE = 2048
CSR_KEY_NAME = 'certkey.key'

# request related
REQ_POLL_INTERVAL = 5.0            # seconds
REQ_POLL_RETRY_COUNT: int = 24

# backup related
BAK_TIME_FMT = '%Y_%m_%d_%H_%M_%S'
BAK_DEFAULT_PATTERN = 'bak_{date_time}.zip'

# logging
LOG_LEVEL = 20    # INFO
LOG_FMT = '[{asctime!s}][{name!s}][{levelname!s}] {msg}'
LOG_DEBUG_FMT = '[{asctime!s}][{name!s}][{levelname!s}] {funcName!s}: {msg}'


# type:ignore[overload]
from typing import Any, Callable, Dict, Optional, overload, Type
import base64

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, ed25519, padding, rsa
from cryptography.hazmat.primitives.asymmetric import utils
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from pyacme.base import _JWSBase, _JWKBase


__all__ = ['JWKRSA', 'JWKES256']


# combine jws module and jwk module to prevent circular import 

# pebble supported alg: 
# https://github.com/letsencrypt/pebble/blob/8a79a257176b5dd587242fb5f524213abc5fd1a7/wfe/jose.go#L17


class _JWSRS(_JWSBase):
    
    alg = ''
    hash_method = ''
    
    def __init__(self, 
                 url: str, 
                 nonce: str, 
                 jwk: 'JWKRSA', 
                 kid: str = '', 
                 payload: Dict[str, Any] = dict()):
        if not isinstance(jwk, JWKRSA):
            raise TypeError(
                f'jwk type "{type(jwk)}" not compatible with {self.alg}'
            )
        super().__init__(self.alg, url, nonce, payload, jwk, kid)
    
    def sign(self, hash_algo: Callable) -> None:
        self.jwk: JWKRSA
        sign_input = self.get_sign_input()
        sig = self.jwk.priv_key.sign(
            data=sign_input,
            # PKCS padding for `RS256` signature
            # see https://tools.ietf.org/html/rfc7518#section-3.3
            padding=padding.PKCS1v15(),
            algorithm=hash_algo()
        )
        self.signature = str(
            base64.urlsafe_b64encode(sig).strip(b'='), encoding='utf-8'
        )
        self.post_body['signature'] = self.signature


class JWSRS256(_JWSRS):

    alg = 'RS256'
    hash_method = 'SHA-256'

    def sign(self) -> None:
        return super().sign(hashes.SHA256)


class JWKRSA(_JWKBase):
    
    kty = 'RSA'
    related_JWS = JWSRS256
    
    def __init__(self, priv_key: rsa.RSAPrivateKey, **kwargs):
        """
        private key is need for RSA JWK to generate signature.
        
        following keyword param must be supplied:
         * n: int
         * e: int
        """
        self.n: int
        self.e: int
        self.priv_key = priv_key
        self.priv_key_path = ''
        # update path of private key
        if 'priv_key_path' in kwargs:
            self.priv_key_path = kwargs['priv_key_path']
        super().__init__(self.kty, **kwargs)
    
    def _check_kty_param(self, kwargs: dict) -> None:
        # TODO check type of n, e
        if not "n" in kwargs:
            raise TypeError(f'missing param "n" for key type {self.kty}')
        if not "e" in kwargs:
            raise TypeError(f'missing param "e" for key type {self.kty}')
    
    def _update_container(self) -> None:
        self._container['kty'] = self.kty
        self._container['n'] = self._b64_encode_int(self.n)
        self._container['e'] = self._b64_encode_int(self.e)


# ES jwk


class _JWSES(_JWSBase):

    alg = ''
    hash_method = ''

    def __init__(self, 
                 url: str, 
                 nonce: str, 
                 jwk: '_JWKESBase', 
                 kid: str, 
                 payload: Dict[str, Any],
                 jwk_target_type: Type):
        if not isinstance(jwk, jwk_target_type):
            raise TypeError(
                f'jwk type "{type(jwk)}" not compatible with {self.alg}'
            )
        super().__init__(self.alg, url, nonce, payload, jwk, kid)
    
    def sign(self, hash_algo: Callable, r_size: int, s_size: int) -> None:
        self.jwk: _JWKESBase
        sign_input = self.get_sign_input()

        # the sign() is asn.1 encoded, does not compatible with rfc7515/7518
        sig = self.jwk.priv_key.sign(
            data=sign_input,
            signature_algorithm=ec.ECDSA(hash_algo())
        )
        # https://tools.ietf.org/html/rfc7515#appendix-A.3.1
        # directly use r and s to construct the signature
        r, s = utils.decode_dss_signature(sig)
        r_byte, s_byte = r.to_bytes(r_size, 'big'), s.to_bytes(s_size, 'big')
        r_s = r_byte + s_byte

        self.signature = str(
            base64.urlsafe_b64encode(r_s).strip(b'='), encoding='utf-8'
        )
        self.post_body['signature'] = self.signature


class _JWKESBase(_JWKBase):

    kty = 'EC'
    related_JWS = _JWSES

    def __init__(self, priv_key: ec.EllipticCurvePrivateKey, **kwargs):
        """
        see https://tools.ietf.org/html/rfc7518#section-6.2

        following keyword param must be supplied:
         * x: int
         * y: int
        """
        self.x: int
        self.y: int
        self.priv_key = priv_key
        self.crv = kwargs['crv']
        super().__init__(self.kty, **kwargs)
    
    def _check_kty_param(self, kwargs: dict) -> None:
        if not "x" in kwargs:
            raise TypeError(f'missing param "x" for key type {self.kty}')
        if not "y" in kwargs:
            raise TypeError(f'missing param "y" for key type {self.kty}')

    def _update_container(self) -> None:
        self._container['kty'] = self.kty
        self._container['x'] = self._b64_encode_int(self.x)
        self._container['y'] = self._b64_encode_int(self.y)
        self._container['crv'] = self.crv


class JWSES256(_JWSES):

    alg = 'ES256'
    hash_method = 'SHA-256'

    def __init__(self, 
                 url: str, 
                 nonce: str, 
                 jwk: 'JWKES256', 
                 kid: str = '', 
                 payload: Dict[str, Any] = dict()):
        super().__init__(url, nonce, jwk, kid, payload, JWKES256)
    
    def sign(self) -> None:
        return super().sign(hashes.SHA256, 32, 32)


class JWKES256(_JWKESBase):

    related_JWS = JWSES256

    def __init__(self, priv_key: ec.EllipticCurvePrivateKey, **kwargs):
        x = priv_key.public_key().public_numbers().x
        y = priv_key.public_key().public_numbers().y
        super().__init__(priv_key, x=x, y=y, crv='P-256', **kwargs)




from typing import Any, Dict, List, Sequence
from pathlib import Path
from multiprocessing import Process
import time
import argparse
import logging

from pyacme.util import generate_privkey, get_keyAuthorization, \
                        main_param_parser, run_http_server, jwk_factory
from pyacme.ACMEobj import ACMEAccount, ACMEAuthorization, ACMEOrder
from pyacme.actions import ACMEAccountActions
from pyacme.request import ACMERequestActions
from pyacme.dnsprovider.dispatch import DNS01ChallengeRespondHandler
from pyacme.settings import *


logger = logging.getLogger(__name__)
debug = logger.debug
info = logger.info


def wait_for_server_stop(p: Process) -> None:
    # after p is terminate()
    p.join()
    while True:
        if not p.is_alive():
            break
        time.sleep(0.5)
    info(f'server on process {p.pid} stopped')


def http_chall(order_obj: ACMEOrder, 
               chall_path: str) -> List[ACMEAuthorization]:
    """
    create http-01 respond file in arg `chall_path`
    """
    base_path = Path(chall_path).absolute() / '.well-known' / 'acme-challenge'
    for auth in order_obj.auth_objs:
        if auth.chall_http.status == 'valid':
            info(f'challenge for {auth.identifier_value} is already valid')
            continue
        # create repond text for each auth http challenge
        chall_text = get_keyAuthorization(
            token=auth.chall_http.token,
            jwk=order_obj.related_acct.jwk_obj
        )
        with open(base_path/auth.chall_http.token, 'w') as f:
            f.write(chall_text)
        auth.chall_http.respond()
        info(f'respond to http challenge for {auth.identifier_value}')
    return order_obj.auth_objs


def main_finalize(order: ACMEOrder, 
                  subject_names: Dict[str, str], 
                  cert_path: str, 
                  csr_priv_key_type: str,
                  csr_priv_key_size: int):
    order.poll_order_state()
    if order.status == 'ready':
        # if csr_priv_key_type.lower() == 'rsa':
        #     csr_privkey = generate_rsa_privkey(
        #         privkey_dir=cert_path,
        #         keysize=csr_priv_key_size,
        #         key_name=CSR_KEY_NAME
        #     )
        # else:
        #     raise ValueError(
        #         f'not supported csr key type {csr_priv_key_type}'
        #     )
        csr_privkey = generate_privkey(
            key_type=csr_priv_key_type,
            privkey_dir=cert_path,
            key_name=CSR_KEY_NAME,
            key_size=csr_priv_key_size
        )
        order.finalize_order(
            privkey=csr_privkey,
            **subject_names
        )
        debug(str(order))
        info('order finalized')
    else:
        raise ValueError(f'order state "{order.status}" != "ready"')


def main_poll_order_state(auths: List[ACMEAuthorization], 
                          poll_interval: float, 
                          poll_retry_count: int):
    # loop and poll the order state
    while poll_retry_count > 0:
        info('polling for authorization ...')
        for auth in auths:
            auth.poll_auth_state()
            if auth.status == 'invalid':
                raise ValueError(
                    f'authorization for domain {auth.identifier_value} invalid'
                )
            if auth.status != 'valid':
                break
        else:
            # here all auth valid, stop server
            break
        poll_retry_count -= 1
        time.sleep(poll_interval)


def main_download_cert(order: ACMEOrder, cert_path):
    order.poll_order_state()
    if order.status == 'valid':
        order.download_certificate(cert_path)
        debug(str(order))
        info(f'certificates download to {cert_path}')
    else:
        raise ValueError(f'order state "{order.status}" != "valid"')


def main_add_args(args: Sequence = []) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description='A simple acme client written in python'
    )
    parser.add_argument(
        '-d',
        '--domain',
        required=True,
        action='append',
        help='Required, FDQN, international domain should use punycode; '
             'use multiple `-d` to provide more than one domains.'
    )
    parser.add_argument(
        '-c',
        '--contact',
        action='append',
        help="input domain holder's email address for CA to send "
             "notification, use multiple `-c` to provide more than one "
             "contact email."
    )
    parser.add_argument(
        '-C',
        '--country_code',
        required=True,
        help='Required, two-digit country code, e.g. CN'
    )
    parser.add_argument(
        '--csr_subjects',
        help='key=value string to csr values besisdes C and CN, '
             'e.g. "ST=State,L=Locality,O=test Org,emailAddres=test@email.org"'
    )
    parser.add_argument(
        '--account_private_key',
        help='absolute path to a pem private key file. '
             'RSA key size must be larger than 2048 and multiple of 4'
    )
    parser.add_argument(
        '--not_before',
        help='a date time string, acme order will not be availabe '
             'before this time'
    )
    parser.add_argument(
        '--not_after',
        help='a date time string, acme order will not be availabe '
             'after this time'
    )
    parser.add_argument(
        '-w',
        '--working_directory',
        default=WD_DEFAULT,
        help=f'dafault is {WD_DEFAULT} ; '
             f'cert can be found at working_directroy/{WD_CERT}'
    )
    parser.add_argument(
        '-m',
        '--mode',
        choices=MODES_SUPPORTED,
        default=MODES_DEFAULT,
        help='decide how to complete acme challenge, '
             f'default "{MODES_DEFAULT}"; '
             'root privilege needed for "http" mode'
    )
    parser.add_argument(
        '--dnsprovider',
        choices=DNS_PROVIDERS,
        default=DNS_DEFAULT_PROVIDER,
        help='select one dnsprovider, '
             f'current support following providers {DNS_PROVIDERS}, '
             f'default provider {DNS_DEFAULT_PROVIDER}'
    )
    parser.add_argument(
        '-k',
        '--access_key',
        help='access key or token to dns provider, if mode is "dns", this '
             'option is required; if mode is "http", this option is omitted'
    )
    parser.add_argument(
        '-s',
        '--secret',
        help='secret or token to dns provider, if mode is "dns", and '
             'dnsprovider is "aliyun" this option is required; '
             'if mode is "http", this option is omitted'
    )
    parser.add_argument(
        '--dns_specifics',
        help='for certain dnsproviders, pass string like '
             '"key1=value1,key2=value2 ..."'
    )
    parser.add_argument(
        '--CA_entry',
        default=LETSENCRYPT_PRODUCTION,
        help=f'url to a CA /directory, default is {LETSENCRYPT_PRODUCTION}'
    )
    parser.add_argument(
        '--poll_interval',
        type=float,
        default=REQ_POLL_INTERVAL,
        help='Optional, seconds between each authorization poll, '
             f'default {REQ_POLL_INTERVAL}'
    )
    parser.add_argument(
        '--poll_retry_count',
        type=int,
        default=REQ_POLL_RETRY_COUNT,
        help='total count of authorization poll retry, '
             f'default {REQ_POLL_RETRY_COUNT}'
    )
    parser.add_argument(
        '--csr_priv_key_type',
        # choices=['rsa'],
        choices=CSR_SUPPORTED_KEY_TYPE,
        default=CSR_DEFAULT_KEY_TYPE,
        help='select key type to sign CSR, default "rsa"'
    )
    parser.add_argument(
        '--csr_priv_key_size',
        type=int,
        default=2048,
        help='Optional, key size of key that will sign CSR, default 2048'
    )
    parser.add_argument(
        '--chall_resp_server_port',
        type=int,
        default=HTTP_SERVING_PORT,
        help='the port used when responding to http-01 challenge; '
             f'default on port {HTTP_SERVING_PORT}, root previlige needed'
    )
    parser.add_argument(
        '--no_ssl_verify',
        action='store_true',
        help='disable ssl certificate verification when requesting acme '
             'resources, default False'
    )
    parser.add_argument(
        '--acct_key_type',
        choices=KEY_ACCT_KEYTYPE,
        default=KEY_ACCT_KEYTYPE_DEFAULT,
        help='decide acme server private key type, default '
             f'{KEY_ACCT_KEYTYPE_DEFAULT}'
    )
    parser.add_argument(
        '--debug',
        action='store_true',
        help='set this option to output debug message'
    )

    from pyacme import __version__
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s ' + f'{__version__}'
    )
    if args:
        parsed_args = parser.parse_args(args)
    else:
        parsed_args = parser.parse_args()
    return parsed_args


def main(*,
         domains: List[str], 
         contact: List[str],
         acct_priv_key: str, 
         not_before: str,
         not_after: str,
         subject_names: Dict[str, str],
         cert_path: str, 
         chall_path: str, 
         mode: str,
         dnsprovider: str,
         access_key: str,
         secret: str,
         dns_specifics: Dict[str, Any],
         CA_entry: str,
         poll_interval: float,
         poll_retry_count: int,
         csr_priv_key_type: str,
         csr_priv_key_size: int,
         chall_resp_server_port: int,
         no_ssl_verify: bool) -> None:

    if no_ssl_verify:
        ACMERequestActions.verify = False

    # set url for CA 
    ACMERequestActions.set_directory_url(CA_entry)
    ACMERequestActions.query_dir()
    req = ACMERequestActions()
    req.new_nonce()
    # init key object
    jwk = jwk_factory(acct_priv_key)
    # init acct action
    acct_action = ACMEAccountActions(req)
    acct = ACMEAccount.init_by_create(
        jwk=jwk,
        acct_actions=acct_action,
        contact=contact
    )
    debug(str(acct))
    if acct._resp.status_code == 200:
        info('account existed and fetched')
    elif acct._resp.status_code == 201:
        info('new account created')
    # create new order for domains
    order = acct.new_order(
        identifiers=domains,
        not_after=not_after,
        not_before=not_before
    )
    debug(str(order))
    info(f'order created {domains}')


    if mode == 'http':
        # start http server
        server_p = Process(
            target=run_http_server,
            args=(chall_path, chall_resp_server_port),
            # daemon=True
        )
        server_p.start()

        # TODO 
        # waiting is needed when pytest-cov is enabled, reason unknown
        time.sleep(1)

        try:
            auths = http_chall(order, chall_path=chall_path)
            for a in auths: debug(str(a))
            info('http challenge responded')

            # loop and poll the order state
            main_poll_order_state(auths, poll_interval, poll_retry_count)

            # do not stop server in `for else` above to avoid deadlock
            info('all authorizaitons valid, stopping server')
            server_p.terminate()
            
            # finalize order
            main_finalize(
                order, subject_names, cert_path, 
                csr_priv_key_type, csr_priv_key_size
            )
            main_download_cert(order, cert_path)
            wait_for_server_stop(server_p)

            info('http mode all done')
        except Exception as e:
            logger.warning('stopping server due to exception')
            server_p.terminate()
            wait_for_server_stop(server_p)
            raise e
    
    elif mode == 'dns':
        handler = DNS01ChallengeRespondHandler(
            order_obj=order,
            dnsprovider=dnsprovider,
            access_key=access_key,
            secret=secret,
            **dns_specifics
        )
        try:
            auths = handler.dns_chall_respond()
            info('dns challenge responded')
            # loop and poll the order state
            main_poll_order_state(auths, poll_interval, poll_retry_count)
            info('all authorizaitons valid, clearing dns record')
            handler.clear_dns_record()
            # finalize order
            main_finalize(
                order, subject_names, cert_path, 
                csr_priv_key_type, csr_priv_key_size
            )
            main_download_cert(order, cert_path)
            
            info('dns mode all done')
        except Exception as e:
            logger.warning('removing dns record due to exception')
            handler.clear_dns_record()
            raise e
    else:
        raise ValueError(f'not supported mode {mode}')


def main_entry_point():
    args = main_add_args()
    debug(args)
    param_dict = main_param_parser(args)
    debug(param_dict)
    main(**param_dict)


def main_entry_point_test(p: Sequence):
    args = main_add_args(p)
    debug(args)
    param_dict = main_param_parser(args)
    debug(param_dict)
    main(**param_dict)


import os
from typing import List, Tuple, Union, Dict, Sequence, Callable
from pathlib import Path
import time
import json
import subprocess

import requests
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend

# for running this file directly
import sys
sys.path.append(str(Path(__file__).parents[1].absolute()))

from pyacme.base import _JWKBase
from pyacme.execution import main_entry_point_test
from pyacme.util import get_keyAuthorization, get_dns_chall_txt_record
from test_settings import *


CERT_DIR = Path('.cert_files')

TEST_CONTACTS = ['mailto:min641366609@live.com']
TEST_CONTACTS_MOD = ['mailto:calvin.cheng@synergyfutures.com']

SLEEP_AFTER_DOCKER = 4

# print(RSA_PUB_1)

global_docker_is_running = False


def _run_cmd(*cmd_args: str) -> None:
    completed = subprocess.run(
        args=cmd_args,
        # capture_output=True,
        stdout=subprocess.PIPE,
        encoding='utf-8',
        # should raise exception if docker failed to run
        check=True
    )
    # if completed.stdout:
    #     print(completed.stdout)
    # # docker output seems to be on stderr
    # print(completed.stderr)


def run_pebble_docker(docker_file_path: str) -> None:
    """run `docker-compose up` on given docker-compose.yml in the background"""
    global global_docker_is_running
    _run_cmd('docker-compose', '-f', docker_file_path, 'up', '-d')
    # wait for a while to make sure container is up and running
    time.sleep(SLEEP_AFTER_DOCKER)
    print('docker container running')
    global_docker_is_running = True


def stop_pebble_docker(*docker_names: str) -> None:
    global global_docker_is_running
    """`docker stop DOCKER_NAMES ...`"""
    _run_cmd('docker', 'stop', *docker_names)
    print('docker container stopped')
    global_docker_is_running = False


def restart_pebble_docker(docker_file_path: str, *docker_names: str) -> None:
    stop_pebble_docker(*docker_names)
    run_pebble_docker(docker_file_path)


def restart_pebble_docker_specific():
    if global_docker_is_running:
        restart_pebble_docker(
            PEBBLE_DOCKER_FILE, 
            PEBBLE_CONTAINER, 
            PEBBLE_CHALLTEST_CONTAINER
        )


def add_dns_A(host: str, addr: str) -> requests.Response:
    """add an A record to pebble-challtestsrv"""
    data = json.dumps({'host': host, 'address': addr})
    resp = requests.post(PEBBLE_CHALLTEST_DNS_A, data=data)
    return resp


def add_dns_01(token: str, domain: str, jwk: _JWKBase) -> requests.Response:
    """
    add dns-01 challenge respond to pebble challtest
    """
    value = get_dns_chall_txt_record(token, jwk)
    domain = f'_acme-challenge.{domain}.'
    resp = requests.post(
        url=PEBBLE_CHALLTEST_DNS01,
        data=json.dumps({'host': domain, 'value': value})
    )
    return resp


def add_http_01(token: str, jwk: _JWKBase) -> requests.Response:
    """
    add repsond to challtestsrv, pebble will send challenge request to 
    challtestsrv if pebble recieve `respond_to_challenge` request
    """
    content = get_keyAuthorization(token, jwk)
    resp = requests.post(
        url=PEBBLE_CHALLTEST_HTTP01,
        data=json.dumps({'token': token, 'content': content})
    )
    return resp


T = List[Tuple[tuple, str]]


def load_test_keys(*key_pair_path: Tuple[str, str]) -> T:
    """
    load pem file into key object
    """
    key_pairs: T = []
    for pub_path, priv_path in key_pair_path:
        with open(pub_path, 'rb') as pub_f:
            pub_key = serialization.load_pem_public_key(
                pub_f.read(),
                backend=default_backend()
            )
        with open(priv_path, 'rb') as priv_f:
            priv_key = serialization.load_pem_private_key(
                priv_f.read(),
                password=None,
                backend=default_backend()
            )
        key_pairs.append(((pub_key, priv_key), priv_path))
    return key_pairs


def download_root_cert(root_cert_path: Path, 
                       url: str = 'https://localhost:15000/roots/0',
                       name = "pebble-root-cert.pem"):
    """download root cert from pebble container"""
    subprocess.run(
        [
            'wget', url, 
            '--no-check-certificate',
            '-O', f'{root_cert_path/name!s}',
            '--quiet'
        ]
    )


def openssl_verify(cert_path: Union[Path, str], 
                   chain_path: Union[Path, str], 
                #    root_cert_path = CERT_DIR,
                   root_cert_path: Union[Path, str],
                   root_cert_name = "pebble-root-cert.pem"):
    """run `openssl verify` on downloaded cert"""
    p = subprocess.run(
        [
            'openssl', 'verify',
            '-CAfile', f'{root_cert_path/root_cert_name!s}',
            '-untrusted', str(chain_path),
            str(cert_path)
        ]
    )
    return p


def add_host_entry(domains: List[str], addr: str) -> None:
    """
    add entry like `127.0.0.1 test.local` to /etc/hosts, if entry exists, skip
    """
    if os.getuid() != 0:
        print(
            f'no root present, did not write to host; ' 
            f'current uid={os.getuid()}'
        )
        return
    check_target = [[addr, d] for d in domains]
    with open('/etc/hosts', 'r') as f:
        lines = [l.strip().split('\t') for l in f.readlines()]
    with open('/etc/hosts', 'a+') as f:
        for addr_d in check_target:
            if addr_d in lines:
                print(f'{addr_d} exists in /etc/hosts')
                continue
            else:
                entry = '\t'.join(addr_d)+'\n'
                f.write(entry)
                print(f'{entry} written to /etc/hosts')


def run_pebble_standalone_container(name: str = 'pebble'):
    """run a pebble container using host's network"""
    p = subprocess.run(
        [
            'docker', 'run', 
            '-d', 
            '-e', '"PEBBLE_VA_NOSLEEP=1"', 
            '--network=host',
            '--name', name,
            'letsencrypt/pebble'
        ],
        check=False,
        # capture_output=True
        stdout=subprocess.PIPE
    )
    if p.returncode != 0:
        raise ValueError(p.stderr.decode('utf-8'))
    time.sleep(SLEEP_AFTER_DOCKER)


def stop_pebble_standalone_container(name: str = 'pebble') -> None:
    subprocess.run(['docker', 'stop', name], check=True)
    # subprocess.run(['docker', 'container', 'prune', '-f'], check=True)


def create_py_http_server(bind: str, port: str, path: str) -> subprocess.Popen:
    # use a dynamic path, in case running test in root
    py_path = sys.executable
    param = [
        '-m', 'http.server', 
        '--bind', bind,
        '--directory', path,
        port
    ]
    try:
        p =  subprocess.Popen([py_path, *param])
    except FileNotFoundError:
        py_path = 'python3'
        p = subprocess.Popen([py_path, *param])
    print(f'server created at pid={p.pid}')
    return p


if __name__ == '__main__':
    print(Path(__file__).parent)


from abc import ABCMeta, abstractmethod


class _DNSProviderBase(metaclass=ABCMeta):

    def __init__(self, access_key: str, secret: str, *args, **kwargs) -> None:
        pass

    @abstractmethod
    def add_txt_record(self, 
                       identifier: str, 
                       value: str, 
                       *args, 
                       **kwargs) -> None:
        pass

    @abstractmethod
    def clear_dns_record(self, *args, **kwargs) -> None:
        pass


from typing import Dict, Any, List, Tuple, Type, TypeVar, Union
import base64
import json
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

import requests

from pyacme.ACMEobj import ACMEAccount, ACMEChallenge, Empty, ACMEOrder
from pyacme.ACMEobj import ACMEAuthorization
from pyacme.exceptions import ACMEError
from pyacme.base import _JWSBase, _JWKBase, _AcctActionBase
from pyacme.util import parse_csr
from pyacme.jws import JWSRS256
from pyacme.jwk import JWKRSA


# TODO cancel param jws_type


__all__ = ['ACMEAccountActions', 'RS256Actions']


TJWS = TypeVar('TJWS', bound=Type[_JWSBase])


class ACMEAccountActions(_AcctActionBase):
    """helper class to execute account management actions"""
    
    def create_acct(self, 
                    jwk: _JWKBase, 
                    contact: List[str], 
                    jws_type: TJWS) -> requests.Response:
        """
        create acme account using a pub key, contact should
        be a list of "mailto:" address; upon success return
        201-created; if account exist, return 200-OK
        """
        jws = jws_type(
            url=self.req_action.acme_dir['newAccount'],
            nonce=str(self.req_action.nonce),
            jwk=jwk,
            payload={
                'termsOfServiceAgreed': True,
                'contact': contact
            }
        )    # type: ignore
        jws.sign()
        resp = self.req_action.new_account(jws)
        if resp.status_code >= 400:
            raise(ACMEError(resp))
        return resp
    
    def query_acct(self, jwk: _JWKBase, jws_type: TJWS) -> requests.Response:
        """
        query if an account is recorded by server, will not 
        create new account; upon success return 200-OK
        
        see https://tools.ietf.org/html/rfc8555#section-7.3.1
        """
        jws = jws_type(
            url=self.req_action.acme_dir['newAccount'],
            nonce=str(self.req_action.nonce),
            jwk=jwk,
            payload={
                'onlyReturnExisting': True
            }
        )    # type: ignore
        jws.sign()
        resp = self.req_action.new_account(jws)
        if resp.status_code >= 400:
            raise(ACMEError(resp))
        return resp
    
    def update_acct(self, 
                    acct_obj: ACMEAccount,
                    jws_type: TJWS,
                    **kwargs) -> requests.Response:
        """
        send updated payload to an account url, update on 
        `termsOfServiceAgreed`, `orders` and `status` will be ignored; 
        `contact` is the usual update target

        see https://tools.ietf.org/html/rfc8555#section-7.3.2
        """
        jws = jws_type(
            url=acct_obj.acct_location,
            nonce=str(self.req_action.nonce),
            payload=kwargs,
            jwk=acct_obj.jwk_obj,
            kid=acct_obj.acct_location
        )    # type: ignore
        jws.sign()
        resp = self.req_action._request(
            url=acct_obj.acct_location,
            method='post',
            jws=jws
        )
        if resp.status_code >= 400:
            raise(ACMEError(resp))
        return resp
    
    def external_acct_binding(self):
        """
        see https://tools.ietf.org/html/rfc8555#section-7.3.4
        """
        pass

    def acct_key_rollover(self, 
                          acct_obj: ACMEAccount,
                          jwk_new: _JWKBase,
                          jws_type: TJWS) -> requests.Response:
        """
        change the public key that is associtated with an account, both new and
        old key should be provided as `jwk` instance.

        see https://tools.ietf.org/html/rfc8555#section-7.3.5
        """
        inner_jws = jwk_new.related_JWS(
            url=self.req_action.acme_dir['keyChange'],
            # nonce ignored in inner jws
            nonce='',
            # inner payload is a "keyChange" object, see rfc8555 page 41
            payload={
                'account': acct_obj.acct_location,
                'oldKey': acct_obj.jwk_obj._container
            },
            jwk=jwk_new
        )    # type: ignore
        inner_jws.protected.pop('nonce')
        inner_jws.sign()

        outer_jws = jws_type(
            url=self.req_action.acme_dir['keyChange'],
            nonce=str(self.req_action.nonce),
            payload=inner_jws.post_body,
            jwk=acct_obj.jwk_obj,
            kid=acct_obj.acct_location
        )    # type: ignore
        outer_jws.sign()
        resp = self.req_action.key_change(outer_jws)
        if resp.status_code >= 400:
            raise ACMEError(resp)
        return resp
    
    def deactivate_acct(self, 
                        acct_obj: ACMEAccount, 
                        jws_type: TJWS) -> requests.Response:
        """
        deactivate an account, issued certificate will not be revoked.
        
        see https://tools.ietf.org/html/rfc8555#section-7.3.6
        """
        jws = jws_type(
            url=acct_obj.acct_location,
            nonce=str(self.req_action.nonce),
            payload={'status': 'deactivated'},
            jwk=acct_obj.jwk_obj,
            kid=acct_obj.acct_location
        )    # type: ignore
        jws.sign()
        resp = self.req_action._request(
            url=acct_obj.acct_location,
            method='post',
            jws=jws
        )
        if resp.status_code >= 400:
            raise(ACMEError(resp))
        return resp
    
    def new_order(self, 
                  acct_obj: ACMEAccount, 
                  identifiers: List[Dict[str, Any]],
                  not_before: str,
                  not_after: str,
                  jws_type: TJWS) -> requests.Response:
        """
        request for new order, expect 201-created upon success. 

         * `identifiers`: e.g. `[{'type': 'dns', 'value': 'test.org'}, ...]` 
         * `not_before`, `not_after`: both optional, datetime string specified 
         in https://tools.ietf.org/html/rfc3339, can be empty `''`
         * return an `requests.Response`;

        see https://tools.ietf.org/html/rfc8555#section-7.4, page 45-46
        """
        payload = {'identifiers': identifiers}
        if not_before:
            payload['notBefore'] = not_before
        if not_after:
            payload['notAfter'] = not_after
        
        jws = jws_type(
            url=self.req_action.acme_dir['newOrder'],
            nonce=str(self.req_action.nonce),
            payload=payload,
            jwk=acct_obj.jwk_obj,
            kid=acct_obj.acct_location
        )
        jws.sign()
        resp = self.req_action.new_order(jws)
        if resp.status_code >= 400:
            raise ACMEError(resp)
        return resp
    
    def post_as_get(self, 
                    url: str,
                    acct_obj: ACMEAccount,
                    jws_type: TJWS) -> requests.Response:
        """
        POST-as-GET to a resource's url, payload is empty string `""`; 
         * return `requests.Response`

        see https://tools.ietf.org/html/rfc8555#section-7.5
        """
        jws = jws_type(
            url=url,
            nonce=str(self.req_action.nonce),
            payload="",
            jwk=acct_obj.jwk_obj,
            kid=acct_obj.acct_location
        )
        jws.sign()
        resp = self.req_action._request(
            url=url,
            method='post',
            jws=jws
        )
        if resp.status_code >= 400:
            raise ACMEError(resp)
        return resp
    
    def respond_to_challenge(self, 
                             chall_obj: ACMEChallenge,
                             jws_type: TJWS) -> requests.Response:
        """
        respond to a challenge url stated in the `challenges` attr in an
        `ACMEAuthorization` instance; payload is empty dict `{}`; expect 200-OK
        if chanllenge object is updated by server.
         * return `requests.Response`

        see https://tools.ietf.org/html/rfc8555#section-7.5.1
        """
        jws = jws_type(
            # the "url" attr of chall_obj, not the location of the obj
            url=chall_obj.url,
            nonce=str(self.req_action.nonce),
            payload=dict(),
            jwk=chall_obj.related_auth.related_order.related_acct.jwk_obj,
            kid=chall_obj.related_auth.related_order.related_acct.acct_location
        )
        jws.sign()
        resp = self.req_action._request(
            url=chall_obj.url,
            method='post',
            jws=jws
        )
        if resp.status_code >= 400:
            raise ACMEError(resp)
        return resp
    
    def deactivate_auth(self, 
                        auth_obj: ACMEAuthorization,
                        jws_type: TJWS) -> requests.Response:
        """
        request to deactivate an authorization; `auth_obj` should be one of the
        element from `acct_obj.auth_objs`; accot_obj will be updated
         * payload `{"status": "deactivated"}`
         * return `requests.Response`

        see https://tools.ietf.org/html/rfc8555#section-7.5.2
        """
        jws = jws_type(
            url=auth_obj.auth_location,
            nonce=str(self.req_action.nonce),
            payload={'status': 'deactivated'},
            jwk=auth_obj.related_order.related_acct.jwk_obj,
            kid=auth_obj.related_order.related_acct.acct_location
        )
        jws.sign()
        resp = self.req_action._request(
            url=auth_obj.auth_location,
            method='post',
            jws=jws
        )
        if resp.status_code >= 400:
            raise ACMEError(resp)
        return resp

    def finalize_order(self, 
                       order_obj: ACMEOrder,
                       privkey: Union[RSAPrivateKey, str],
                       domains: List[str],
                       subject_names: Dict[str, str],
                       jws_type: TJWS) -> requests.Response:
        """
        request to finalize acme order. `ACMEOrder` is tied to one 
        `AMCEAccount`, expect 200-OK if finalize is completed.
         * payload is b64 encoded `CSR`
         * return `requests.Response`

        see https://tools.ietf.org/html/rfc8555#section-7.4 p47
        """
        csr_der_output = parse_csr(
            privkey=privkey,
            domains=domains,
            **subject_names
        )
        csr_der_b = base64.urlsafe_b64encode(csr_der_output).strip(b'=')
        jws = jws_type(
            url=order_obj.finalize,
            nonce=str(self.req_action.nonce),
            payload={
                'csr': csr_der_b.decode('utf-8'),
            },
            jwk=order_obj.related_acct.jwk_obj,
            kid=order_obj.related_acct.acct_location
        )
        jws.sign()
        resp = self.req_action._request(
            url=order_obj.finalize,
            method='post',
            jws=jws
        )
        if resp.status_code >= 400:
            raise ACMEError(resp)
        return resp



from typing import Dict, Any
import json
import logging

import requests

from pyacme.settings import LETSENCRYPT_STAGING
from pyacme.base import _JWSBase, _ACMERequestBase
from pyacme.exceptions import ACMEError


logger = logging.getLogger(__name__)
debug = logger.debug


class Nonce:
    """
    Replay-Nonce http header, it can be get from /newNonce or
    returned by the latest response header;
    see https://tools.ietf.org/html/rfc8555#section-7.2
    """
    
    def __init__(self, nonce: str = ''):
        self.latest = nonce
    
    def update(self, nonce: str) -> None:
        self.latest = nonce
    
    def update_from_resp(self, resp: requests.Response) -> None:
        self.latest = resp.headers['Replay-Nonce']
    
    def __str__(self):
        return self.latest
    
    __repr__ = __str__


class ACMERequestActions(_ACMERequestBase):
    
    # TODO possible threadsafe solution
    # for nonce updating
    common_header = {
        # https://tools.ietf.org/html/rfc8555#section-6.2
        'Content-Type': 'application/jose+json',
        # 'User-Agent': '',
    }
    # use test as default for now
    dir_url = LETSENCRYPT_STAGING
    acme_dir: Dict[str, str] = dict()
    badNonce_max_retry = 5
    verify = True
    
    @classmethod
    def set_directory_url(cls, dir_url: str) -> None:
        cls.dir_url = dir_url
    
    @classmethod
    def query_dir(cls, headers: Dict[str, Any] = dict()) -> None:
        """
        update acme server resources, use GET request
        see https://tools.ietf.org/html/rfc8555#section-7.1
        """
        # no special headers needed
        _resp = requests.get(cls.dir_url, headers=headers, verify=cls.verify)
        cls.acme_dir = json.loads(_resp.text)
        
    def __init__(self, nonce: Nonce = Nonce()):
        self.nonce = nonce
        
    def _request(self, url: str, method: str, jws: _JWSBase, 
                 headers: Dict[str, Any] = dict()) -> requests.Response:
        """send request to arbitrary url with signed jws"""

        def _retry_for_badNonce(e: ACMEError) -> requests.Response:
            # https://tools.ietf.org/html/rfc8555#section-6.5 p15
            # retry automatically if badNonce error response happens
            # replace the old nonce inside jws
            nonlocal jws, self
            jws.update_jws_nonce(str(e.new_nonce))
            jws.sign()
            # new request with retry for new nonce
            resp = getattr(requests, method.lower())(
                url=url,
                data=json.dumps(jws.post_body),
                headers=headers,
                verify=self.verify
            )
            self.nonce.update_from_resp(resp)
            return resp

        headers.update(self.common_header)
        resp: requests.Response
        resp = getattr(requests, method.lower())(
            url=url,
            data=json.dumps(jws.post_body),
            headers=headers,
            verify=self.verify
        )
        self.nonce.update_from_resp(resp)

        # retry multiple times until badNonce clear
        count = 0
        max_count = self.badNonce_max_retry
        while resp.status_code >= 400 and count <= max_count:
            e = ACMEError(resp)
            if e.type == 'badNonce':
                count += 1
                debug(f'badNonce, retry for {count} time')
                resp = _retry_for_badNonce(e)
            else:
                # do not retry if error type is not badNonce
                break

            ## this does not handle <http-date> format, ignore for now
            # if 'Retry-After' in resp.headers:
            #     time.sleep(float(resp.headers['Retry-After']))

        return resp

    def new_nonce(self, headers: Dict[str, Any] = dict()) -> None:
        """get new nonce explicitly, use HEAD method, expect 200"""
        # no special headers needed
        resp = requests.head(
            self.acme_dir['newNonce'], 
            headers=headers,
            verify=self.verify
        )
        self.nonce.update_from_resp(resp)
        
    def new_account(self, jws: _JWSBase, 
                    headers: Dict[str, Any] = dict()) -> requests.Response:
        """
        create new or query existed account according to given publickey.
        if new account is created, expect 201-created; if account existed, 
        expect 200-OK
        
        see https://tools.ietf.org/html/rfc8555#section-7.3
        """
        resp = self._request(
            url=self.acme_dir['newAccount'],
            method='post',
            jws=jws,
            headers=headers
        )
        return resp
    
    def key_change(self, jws: _JWSBase,
                   headers: Dict[str, Any] = dict()) -> requests.Response:
        """
        key rollover request, expect 200-OK on successful change; if new key
        already existed in server with another account, 409-Conflict will 
        return.
        
        see https://tools.ietf.org/html/rfc8555#section-7.3.5 page 43
        """
        resp = self._request(
            url=self.acme_dir['keyChange'],
            method='post',
            jws=jws,
            headers=headers
        )
        return resp

    def new_order(self, jws: _JWSBase,
                  headers: Dict[str, Any] = dict()) -> requests.Response:
        """
        order creation request, expect 201-created if new order is created
        successfully; header `Location` will return, containing an url to the
        created order resources.

        see https://tools.ietf.org/html/rfc8555#section-7.4, page 45-46
        """
        resp = self._request(
            url=self.acme_dir['newOrder'],
            method='post',
            jws=jws,
            headers=headers
        )
        return resp


from typing import Dict, List, Optional, Union
import requests
import json

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey

from pyacme.base import _ACMERespObject, _AcctActionBase, _JWKBase
from pyacme.util import save_cert


__all__ = ['ACMEAccount', 'ACMEOrder', 'ACMEAuthorization', 'ACMEChallenge']


class Empty(_ACMERespObject):
    """represent an empty acme response object"""

    def _update_attr(self, resp: requests.Response, *args, **kwargs) -> None:
        pass


class ACMEAccount(_ACMERespObject):
    """
    An acme account resource, attr `acct_location` is added in addition to
    other rfc specified fields. 
     * pass key-word argument `jwk` to bind a key with the acct_obj
     * pass key-word argument `acct_actions` to bind an `ACMEAccountActions`
     class for an acct_obj

    see https://tools.ietf.org/html/rfc8555#section-7.1.2
    """

    def __init__(self, resp: requests.Response, *args, **kwargs):
        self._order_objs: List['ACMEOrder']
        self.status: str
        self.contact: List[str]
        self.termsOfServiceAgreed: str
        self.externalAccountBinding: Dict[str, str]
        self.orders: str
        super().__init__(resp, *args, **kwargs)

    @classmethod
    def init_by_query(cls, 
                      jwk: _JWKBase, 
                      *, 
                      acct_actions: _AcctActionBase) -> 'ACMEAccount':
        """
        form an `ACMEAccount` instance given a `jwk`, will perform `new_acct` 
        request to server, query if the jwk has an associated account, if not
        then use another request to create one.
        """
        resp = acct_actions.query_acct(jwk, jwk.related_JWS)
        return cls(resp, acct_actions=acct_actions, jwk=jwk)
    
    @classmethod
    def init_by_create(cls, 
                       jwk: _JWKBase,
                       *,
                       acct_actions: _AcctActionBase,
                       contact: List[str]) -> 'ACMEAccount':
        """
        create a new account by given a public key and contact info
        """
        resp = acct_actions.create_acct(
            jwk, contact, jwk.related_JWS
        )
        return cls(resp, acct_actions=acct_actions, jwk=jwk)
    
    def poll_acct_state(self, update_orders: bool = True) -> None:
        """
        update account status by sending post-as-get request to acct location
        """
        if update_orders:
            self.get_orders()
        resp = self.acct_actions.post_as_get(
            url=self.acct_location,
            acct_obj=self,
            jws_type=self.jwk_obj.related_JWS
        )
        self._update_from_resp(resp)
    
    def new_order(self, 
                  *, 
                  identifiers: List[str], 
                  not_before: str = '', 
                  not_after: str = '') -> 'ACMEOrder':
        """
        create new order by giving identifiers
        """
        iden_list = [dict(type='dns', value=i) for i in identifiers]
        resp = self.acct_actions.new_order(
            acct_obj=self,
            identifiers=iden_list,
            not_before=not_before,
            not_after=not_after,
            jws_type=self.jwk_obj.related_JWS
        )
        order_obj = ACMEOrder(resp, related_acct=self)
        self._order_objs.append(order_obj)
        return order_obj
    
    def get_orders(self) -> List['ACMEOrder']:
        """
        iterate through `orders` field of the acct_obj and fill `order_objs`, 
        only orders with state `"pending"` and `"valid"` will return; 
         * may need multiple requests to fetch poll orders, use 
         `requests.Response.links` to handle possible pagination

        see https://tools.ietf.org/html/rfc8555#section-7.1.2.1 Orders List
        """

        def _post_as_get(url: str, order = False) -> requests.Response:
            resp = self.acct_actions.post_as_get(
                url=url, 
                acct_obj=self,
                jws_type=self.jwk_obj.related_JWS
            )
            if order:
                # handling response from post-as-get to an order url;
                # order location may not appear in an non-creation response,
                # add the location back for ACMEOrder constructor
                if not 'Location' in resp.headers:
                    resp.headers['Location'] = url
            return resp

        # get Orders List object by sending post-as-get to "orders" url
        orders_resp = _post_as_get(self.orders)
        url_poll: List[str] = json.loads(orders_resp.text)['orders']

        while 'next' in orders_resp.links:
            url_next = orders_resp.links['next']['url']
            orders_resp = _post_as_get(url_next)
            url_poll += json.loads(orders_resp.text)['orders']
        

        for order_url in url_poll:
            for existed_order_obj in self.order_objs:
                # if an order_obj already exists, call poll_order_state() on it
                if existed_order_obj.order_location == order_url:
                    existed_order_obj.poll_order_state()
                    break
            else:
                # add new order_obj to the acct_obj
                resp = _post_as_get(order_url, order=True)
                order_obj = ACMEOrder(resp, related_acct=self)
                self._order_objs.append(order_obj)
        
        return self.order_objs
    
    def update_account(self, *, contact: List[str], **kwargs) -> None:
        """
        update account info, usaually the contact field;
        change or add email contacts
         * `"contact"`: list of string like `["mailto:test@mail.com"]`
        """
        resp = self.acct_actions.update_acct(
            acct_obj=self,
            jws_type=self.jwk_obj.related_JWS,
            **{'contact': contact, **kwargs}
        )
        self.poll_acct_state()
    
    def account_key_rollover(self, jwk_new: _JWKBase) -> None:
        """change account's pubkey to a new one"""
        resp = self.acct_actions.acct_key_rollover(
            acct_obj=self,
            jwk_new=jwk_new,
            jws_type=self.jwk_obj.related_JWS
        )
        self._set_jwk(jwk_new)
        # new jwk must be used to send post-as-get
        self.poll_acct_state()
    
    def deactivate(self) -> None:
        """
        deactivate the account; after deactivation, 401-Unauthorizaed will
        return from server if any POST or POST-as-GET is sent by the acct.
        """
        resp = self.acct_actions.deactivate_acct(
            acct_obj=self,
            jws_type=self.jwk_obj.related_JWS
        )
        self._update_from_resp(resp)
    
    def _update_attr(self, resp: requests.Response, *args, **kwargs):
        # field and type defined by RFC8555
        attrs = [
            # (field_name, default_value if server not provided)
            ('status', ''),
            ('contact', list()),
            ('termsOfServiceAgreed', ''),       # boolean
            ('externalAccountBinding', dict()), # object
            ('orders', '')
        ]
        for attr in attrs:
            if not (attr[0] in self._raw_resp_body):
                # set default value for rfc8555 stated attr
                setattr(self, attr[0], attr[1])

        if not hasattr(self, 'acct_location'):
            # when updating an acct obj itself, if location already exist
            # then do not give initial value
            self.acct_location = ''
        if 'Location' in resp.headers:
            # sometimes server resp header may not include `"Location"` header,
            # when making account update request
            self.acct_location = resp.headers['Location']

        if 'acct_actions' in kwargs:
            self.acct_actions: _AcctActionBase = kwargs['acct_actions']

        if 'jwk' in kwargs:
            self._set_jwk(kwargs['jwk'])
        
        if not hasattr(self, '_order_objs'):
            self._order_objs = list()
        
        # in case server return some attrs which are not included in rfc8555
        self.__dict__.update(self._raw_resp_body)
    
    def _update_from_resp(self, resp: requests.Response) -> None:
        """update the accout instance by accepting new response"""
        self._set_initial(resp)
        self._update_attr(resp)
    
    def _set_jwk(self, jwk: _JWKBase) -> None:
        self._jwk_obj = jwk
    
    @property
    def order_objs(self) -> List['ACMEOrder']:
        return self._order_objs

    @property
    def jwk_obj(self) -> '_JWKBase':
        return self._jwk_obj


class ACMEOrder(_ACMERespObject):
    """
    An acme order object, attr `order_location` is added in addition to other
    rfc specified fields.
     * pass key-word argument `related_acct` to bind order_obj to an acct_obj

    see https://tools.ietf.org/html/rfc8555#section-7.1.3
    """

    def __init__(self, resp: requests.Response, *args, **kwargs):
        self.status: str
        self.expires: str
        self.notBefore: str
        self.notAfter: str
        self.error: str
        self.identifiers: List[Dict[str, str]]
        self.authorizations: List[str]
        self.finalize: str
        self.certificate: str
        super().__init__(resp, *args, **kwargs)

    def poll_order_state(self) -> None:
        """
        use POST-as-GET to poll and update the `ACMEOrder` object, lastest
        `"authorizations"` and `"status"` are expected
        """
        act = self.related_acct.acct_actions
        jws_type = self.related_acct.jwk_obj.related_JWS
        resp = act.post_as_get(
            url=self.order_location,
            acct_obj=self.related_acct,
            jws_type=jws_type
        )
        self._update_from_resp(resp)
    
    def finalize_order(self, 
                       privkey: Union[RSAPrivateKey, EllipticCurvePrivateKey], 
                       **subject_names) -> None:
        """
        finalize the `ACMEOrder` order by sending to its `"finalize"` url
        """
        act = self.related_acct.acct_actions
        jws_type = self.related_acct.jwk_obj.related_JWS
        resp = act.finalize_order(
            self, 
            privkey=privkey,
            domains=self.identifier_values,
            subject_names=subject_names,
            jws_type=jws_type
        )
        self._update_from_resp(resp)

    def download_certificate(self, 
                             cert_dir: str,
                             cert_fmt = 'pem-certificate-chain'
                             ) -> requests.Response:
        """
        download certificate for an valid order.
        `cer_fmt` can be one of the following:
         * `pem-certificate-chain`
         * `pkix-cert`
         * `pkcs7-mime`

        see https://tools.ietf.org/html/rfc8555#section-7.4.2
        """
        header = {'Accept': f'application/{cert_fmt}'}
        if cert_fmt == 'pem-certificate-chain':
            resp = self._related_acct.acct_actions.post_as_get(
                url=self.certificate,
                acct_obj=self.related_acct,
                jws_type=self.related_acct.jwk_obj.related_JWS
            )
            resp = save_cert(resp, cert_dir)
            return resp
        # TODO
        # elif cert_fmt == 'pkix-cert':
        #     pass
        # elif cert_fmt == 'pkcs7-mime':
        #     pass
        else:
            raise ValueError(f'unrecognized cert format: {cert_fmt}')


    def _update_attr(self, resp: requests.Response, *args, **kwargs) -> None:
        # field and type defined by RFC8555
        attrs = [
            ('status', ''),             # required
            ('expires', ''),

            # identifiers: list of dict, which must contain "type", "value"
            # [{'type': str, 'value': ''}, ...]
            # see rfc8555 p26
            ('identifiers', list()),    # array of objects

            ('notBefore', ''),
            ('notAfter', ''),
            ('error', ''),

            # contains location for auth objects
            ('authorizations', list()), # required, array of strings

            ('finalize', ''),           # required
            ('certificate', '')
        ]
        for attr in attrs:
            if not (attr[0] in self._raw_resp_body):
                setattr(self, attr[0], attr[1])
        
        if not hasattr(self, 'order_location'):
            self.order_location = ''
        if 'Location' in resp.headers:
            self.order_location = resp.headers['Location']

        if 'related_acct' in kwargs:
            self._set_related_acct(kwargs['related_acct'])

        self.__dict__.update(self._raw_resp_body)
        
        self.identifier_values = [i['value'] for i in self.identifiers]

        self._fetch_auth()
    
    def _fetch_auth(self) -> None:
        """when updating order, auth_objs of this order will all be updated"""
        act = self.related_acct.acct_actions
        self._auth_objs: List['ACMEAuthorization'] = []
        for auth_url in self.authorizations:
            resp = act.post_as_get(
                url=auth_url,
                acct_obj=self.related_acct,
                jws_type=self.related_acct.jwk_obj.related_JWS
            )
            self._auth_objs.append(
                ACMEAuthorization(
                    resp=resp, 
                    related_order=self,
                    auth_url=auth_url
                )
             )
    
    def _update_from_resp(self, resp: requests.Response) -> None:
        self._set_initial(resp)
        self._update_attr(resp)

    @property
    def auth_objs(self) -> List['ACMEAuthorization']:
        if not hasattr(self, '_auth_objs'):
            return list()
        return self._auth_objs

    def _set_related_acct(self, acct_obj: ACMEAccount) -> None:
        self._related_acct = acct_obj
    
    @property
    def related_acct(self) -> ACMEAccount:
        return self._related_acct


class ACMEAuthorization(_ACMERespObject):
    """
    An acme authorization object, attr `auth_location` is added in addition 
    to other rfc specified fields.
     * pass key-word argument `related_order` to bind auth_obj to an order_obj
     * pass key-word argument `auth_url` to set this auth_obj's location

    see https://tools.ietf.org/html/rfc8555#section-7.1.4
    """
    def __init__(self, resp: requests.Response, *args, **kwargs) -> None:
        self.status: str
        self.expires: str
        self.identifier: Dict[str, str]
        self.challenges: List[str]
        self.wildcard: str
        super().__init__(resp, *args, **kwargs)

    def poll_auth_state(self) -> None:
        """
        use POST-as-GET to poll and update the `ACMEAuthorization` object,
        `ACMEChallenge` object related to auth will be updated as well
        """
        act = self.related_order.related_acct.acct_actions
        jws_type = self.related_order.related_acct.jwk_obj.related_JWS
        resp = act.post_as_get(
            url=self.auth_location,
            acct_obj=self.related_order.related_acct,
            jws_type=jws_type
        )
        self._update_from_resp(resp)
        # TODO status become deactivated, other fields will be empty
    
    def deactivate_auth(self) -> None:
        """
        deactivate one `ACMEAuthorization` object

        see https://tools.ietf.org/html/rfc8555#section-7.5.2
        """
        act = self.related_order.related_acct.acct_actions
        jws_type = self.related_order.related_acct.jwk_obj.related_JWS
        resp = act.deactivate_auth(self, jws_type=jws_type)
        self._update_from_resp(resp)

    def _update_attr(self, resp: requests.Response, *args, **kwargs) -> None:
        # field and type defined by RFC8555
        attrs = [
            ('status', ''),             # required
            ('expires', ''),

            # identifier: dict, which must contain "type", "value"
            # {'type': str, 'value': ''}
            # see rfc8555 p29
            ('identifier', dict()),     # object

            # challenge: list of dict, may be decided by server
            ('challenges', list()),     # required, array of objects

            ('wildcard', '')            # boolean
        ]
        for attr in attrs:
            if not (attr[0] in self._raw_resp_body):
                setattr(self, attr[0], attr[1])
        
        if not hasattr(self, 'auth_location'):
            self.auth_location = ''
        # auth location comes from order object's authorization field
        if 'auth_url' in kwargs:
            self.auth_location = kwargs['auth_url']

        if 'related_order' in kwargs:
            self._set_related_order(kwargs['related_order'])

        self.__dict__.update(self._raw_resp_body)

        self.identifier_value = self.identifier['value']

        self._set_chall_objs()
    
    def _set_chall_objs(self) -> None:
        # add challenge object
        _resp_body = self._raw_resp_body
        self._chall_objs: List['ACMEChallenge'] = []
        if 'challenges' in _resp_body:
            for chall_dict in _resp_body['challenges']:
                self._chall_objs.append(
                    ACMEChallenge(
                        chall_dict=chall_dict,
                        related_auth=self
                    )
                )

        # set chall_obj for each type
        for chall_obj in self._chall_objs:
            if 'http' in chall_obj.type.lower():
                self.chall_http = chall_obj
            elif 'dns' in chall_obj.type.lower():
                self.chall_dns = chall_obj
            elif 'tls-alpn' in chall_obj.type.lower():
                self.chall_tls_alpn = chall_obj
    
    @property
    def chall_objs(self) -> List['ACMEChallenge']:
        if not hasattr(self, '_chall_objs'):
            return list()
        return self._chall_objs
    
    def _update_from_resp(self, resp: requests.Response, **kwargs) -> None:
        """update the auth instance by accepting new response"""
        self._set_initial(resp)
        self._update_attr(resp, **kwargs)

    def _set_related_order(self, order_obj: ACMEOrder) -> None:
        self._related_order = order_obj
    
    @property
    def related_order(self) -> ACMEOrder:
        return self._related_order


class ACMEChallenge(_ACMERespObject):
    """
    rfc8555 provides no method to poll `ACMEChallenge`, poll related
    `ACMEAuthorization` object to get latest challenge.
     * pass key-word argument `related_auth` to bind chall_obj to an auth_obj

    see https://tools.ietf.org/html/rfc8555#section-8
    """
    def __init__(self, 
                 *, 
                 resp: Optional[requests.Response] = None, 
                 chall_dict: Dict[str, str] = dict(),
                 **kwargs) -> None:
        # https://tools.ietf.org/html/rfc8555#section-7.5.1 p55
        # chall object may be updated by server and returned as response when
        # client responded to a Challenge;
        self.type: str
        self.url: str
        self.status: str
        self.token: str
        self.validated: str
        self.error: str
        if isinstance(resp, requests.Response):
            self._set_initial(resp)
            self._update_attr(resp, **kwargs)
        if chall_dict:
            self._init_by_dict(chall_dict, **kwargs)
    
    def respond(self) -> 'ACMEChallenge':
        """
        respond to this challenge's `"url"` attribute
        """
        acct_obj = self.related_auth.related_order.related_acct
        act = acct_obj.acct_actions
        jws_type = acct_obj.jwk_obj.related_JWS
        resp = act.respond_to_challenge(self, jws_type=jws_type)
        return type(self)(resp=resp)
    
    def _init_by_dict(self, chall_dict: Dict[str, str], **kwargs) -> None:
        self.type = ''
        self.url = ''
        self.status = ''
        self.token = ''
        self.validated = ''
        self.error = ''
        self.__dict__.update(chall_dict)
        if 'related_auth' in kwargs:
            self._set_related_auth(kwargs['related_auth'])
    
    def _update_attr(self, resp: requests.Response, *args, **kwargs) -> None:
        attrs = [
            # required below
            ('type', ''),
            ('url', ''),
            ('status', ''),
            # optional below
            ('token', ''),
            ('validated', ''),
            ('error', '')
        ]
        for attr in attrs:
            if not (attr[0] in self._raw_resp_body):
                setattr(self, attr[0], attr[1])
        
        if 'related_auth' in kwargs:
            self._set_related_auth(kwargs['related_auth'])

        self.__dict__.update(self._raw_resp_body)
    
    def _set_related_auth(self, auth_obj: ACMEAuthorization) -> None:
        self._related_auth = auth_obj
    
    @property
    def related_auth(self) -> ACMEAuthorization:
        return self._related_auth


from typing import Dict, Any, Optional, Type, TypeVar, Union, List
import base64
import json
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey

import requests


class _JWKBase: 
    """JWK object is a memeber of protected header and then encoded by b64"""

    related_JWS: Type['_JWSBase']
    
    def __init__(self, kty: str, **kwargs):
        # _container will be serialised by json then b64 encoded
        self._container: Dict[str, Any] = dict()
        self.kty = kty
        self._check_kty_param(kwargs)
        self.__dict__.update(kwargs)
        self._update_container()
    
    # TODO abstractmethod
    def _check_kty_param(self, kwargs: dict) -> None:
        raise NotImplementedError

    def _update_container(self) -> None:
        raise NotImplementedError
    
    @staticmethod
    def _b64_encode_int(i: int) -> str:
        """
        encode an int using urlsafe base64; hex the int to get string form and
        then transform into bytes for b64 encoding
        """
        hexed = hex(i)[2:]
        if len(hexed) % 2 != 0:
            hexed = '0' + hexed
        i_bytes = bytes.fromhex(hexed)
        b64_encoded = base64.urlsafe_b64encode(i_bytes).strip(b'=')
        return str(b64_encoded, encoding='utf-8')
    
    def __str__(self):
        return str(self._container)
    
    __repr__ = __str__


class _JWSBase:
    
    def __init__(self, 
                 alg: str, 
                 url: str, 
                 nonce: str, 
                 payload: Union[Dict[str, Any], str],
                 jwk: _JWKBase,  
                 kid: str = ''):
        self.alg = alg

        # jwk and kid are exclusive, one of them must be provided
        # see https://tools.ietf.org/html/rfc8555#section-6.2, page 12

        # always provide a `jwk` for now, for storing private key; if `kid` is
        # presented, use `kid` for protected header
        self.jwk = jwk
        self.kid = kid

        self.protected: Dict[str, Any] = {
            'url': url,
            'nonce': nonce,
            'alg': alg,
        }
        if kid:
            self.protected['kid'] = kid
        else:
            self.protected['jwk'] = jwk._container

        self.url = url
        self.nonce = nonce
        # paylaod may be empty `dict()` or empty string `""`
        # see https://tools.ietf.org/html/rfc8555#section-7.5.1 for empty {}
        self.payload = payload
        self.signature = ''
        self.post_body: Dict[str, str] = dict()
    
    def update_jws_nonce(self, new_nonce: str) -> None:
        self.nonce = new_nonce
        self.protected['nonce'] = new_nonce
    
    def get_sign_input(self) -> bytes:
        """
        see https://tools.ietf.org/html/rfc7515#section-2 Signing Input
        """
        protected_json = json.dumps(self.protected)
        protected_b64 = base64.urlsafe_b64encode(
            bytes(protected_json, encoding='utf-8')).strip(b'=')

        # empty payload, see rfc8555 p54 POST example
        # this should not be the literal b'""'
        payload_b64 = b""

        if isinstance(self.payload, (dict, list)):
            # dict or list like payload, empty dict will be json serialized
            payload_json = json.dumps(self.payload)
            payload_b64 = base64.urlsafe_b64encode(
                bytes(payload_json, encoding='utf-8')).strip(b'=')
        elif isinstance(self.payload, str) and bool(self.payload):
            # non-empty string payload
            # see https://tools.ietf.org/html/rfc8555#section-6.3
            payload_b64 = base64.urlsafe_b64encode(
                bytes(self.payload, encoding='utf-8')).strip(b'=')

        self.sign_input = protected_b64 + b'.' + payload_b64

        self.post_body['protected'] = str(protected_b64, encoding='utf-8')
        self.post_body['payload'] = str(payload_b64, encoding='utf-8')
        return self.sign_input
    
    def sign(self, *args, **kwargs) -> None:
        # update signature to self.post_body, str type signature generated
        raise NotImplementedError


class _ACMERespObject:
    """represent an object returned by an acme server"""
    
    def __init__(self, resp: requests.Response, *args, **kwargs):
        # some response may not have content
        self._set_initial(resp)
        self._update_attr(resp, *args, **kwargs)
        # set values for server specified fields that are not in rfc
        # self.__dict__.update(self._raw_resp_body)
    
    def _set_initial(self, resp: requests.Response) -> None:
        self._raw_resp_body = dict()
        if resp.text:
            self._raw_resp_body = json.loads(resp.text)
        self._resp = resp
    
    def _update_attr(self, resp: requests.Response, *args, **kwargs) -> None:
        raise NotImplementedError
    
    def __str__(self):
        # make a copy to prevent changes to origin dict
        cls = type(self).__name__
        _dict = {
            k : v for (k, v) in self.__dict__.items() if not k.startswith('_')
        }
        s = f'{cls}({str(_dict)})'
        return s
    
    __repr__ = __str__


TJWS = TypeVar('TJWS', bound=Type[_JWSBase])


class _ACMERequestBase:

    def __init__(self, nonce):
        """nonce should have a string form that is able to update with each
        reponse recieved from the server"""
        self.nonce = nonce

    def _request(self, url: str, method: str, jws: _JWSBase, 
                 headers: Dict[str, Any] = dict()) -> requests.Response:
        raise NotImplementedError


class _AcctActionBase:

    def __init__(self, req_action: _ACMERequestBase):
        self.req_action = req_action
    
    def create_acct(self, 
                    jwk: _JWKBase, 
                    contact: List[str], 
                    jws_type: TJWS) -> requests.Response:
        raise NotImplementedError

    def query_acct(self, jwk: _JWKBase, jws_type: TJWS) -> requests.Response:
        raise NotImplementedError

    def update_acct(self, 
                    acct_obj: _ACMERespObject,
                    jws_type: TJWS,
                    **kwargs) -> requests.Response:
        raise NotImplementedError

    def external_acct_binding(self):
        raise NotImplementedError

    def acct_key_rollover(self, 
                          acct_obj: _ACMERespObject,
                          jwk_new: _JWKBase,
                          jws_type: TJWS) -> requests.Response:
        raise NotImplementedError

    def deactivate_acct(self, 
                        acct_obj: _ACMERespObject, 
                        jws_type: TJWS) -> requests.Response:
        raise NotImplementedError

    def new_order(self, 
                  acct_obj: _ACMERespObject, 
                  identifiers: List[Dict[str, Any]],
                  not_before: str,
                  not_after: str,
                  jws_type: TJWS) -> requests.Response:
        raise NotImplementedError

    def post_as_get(self, 
                    url: str,
                    acct_obj: _ACMERespObject,
                    jws_type: TJWS) -> requests.Response:
        raise NotImplementedError

    def respond_to_challenge(self, 
                             chall_obj: _ACMERespObject,
                             jws_type: TJWS) -> requests.Response:
        raise NotImplementedError

    def deactivate_auth(self, 
                        auth_obj: _ACMERespObject,
                        jws_type: TJWS) -> requests.Response:
        raise NotImplementedError

    def finalize_order(self, 
                       order_obj: _ACMERespObject,
                       privkey: Union[RSAPrivateKey, EllipticCurvePrivateKey],
                       domains: List[str],
                       subject_names: Dict[str, str],
                       jws_type: TJWS) -> requests.Response:
        raise NotImplementedError

    def __str__(self) -> str:
        s = f'current_nonce: {str(self.req_action.nonce)}'
        return s
    
    def __repr__(self) -> str:
        return repr(str(self))


"""
Install the following package:
`pip install aliyun-python-sdk-core`
`pip install aliyun-python-sdk-alidns`
"""

import json
import logging
from typing import List

from aliyunsdkcore.client import AcsClient
from aliyunsdkalidns.request.v20150109.AddDomainRecordRequest \
    import AddDomainRecordRequest
from aliyunsdkalidns.request.v20150109.DescribeDomainRecordInfoRequest \
    import DescribeDomainRecordInfoRequest
from aliyunsdkalidns.request.v20150109.DeleteDomainRecordRequest \
    import DeleteDomainRecordRequest

from .dnsproviderbase import _DNSProviderBase


logger = logging.getLogger(__name__)
info = logger.info
debug = logger.debug


def create_client(access_key: str, secret: str, r = 'cn-hangzhou') -> AcsClient:
    """
    provide access key, secret, return client object; key and secret 
    usually from aliyun RAM; region uses default value `'cn-hangzhou'`
    """
    return AcsClient(access_key, secret, r)


def add_dns_txt_record(client: AcsClient, 
                       domain: str, rr: str, value: str) -> dict:
    """
    `domain`: the literal domain string, no punycode used;

    `rr`: fill in `"_acme-challenge"`;

    `value`: fill in dns-01 challenge response text;

    upon success record addition, return `RecordId` like the follwing:
    ```
    {'RequestId': 'EF1FA305-8410-48EC-A599-C0F70B2D66B1',
     'RecordId': '20963244238718976'}
    ```
    the record id will be used for later record removal
    """
    domain_split = domain.split('.')
    if len(domain_split) > 2:
        # handle passed in domain like 'test.xn--jhqy4a5a064kimjf01df8e.host'
        sub, primary = '.'.join(domain_split[:-2]), '.'.join(domain_split[-2:])
        rr = rr + '.' + sub
    else:
        sub = ''
        primary = domain

    debug(f'add dns record rr={rr} sub={sub} primary={primary} value={value}')

    request = AddDomainRecordRequest()
    request.set_accept_format('json')
    request.set_DomainName(primary)
    request.set_RR(rr)
    request.set_Value(value)
    request.set_Type("TXT")
    response = client.do_action_with_exception(request)
    return json.loads(str(response, encoding='utf-8'))


def query_domain_record_by_id(client: AcsClient, record_id: str) -> dict:
    request = DescribeDomainRecordInfoRequest()
    request.set_accept_format('json')
    request.set_RecordId(record_id)
    response = client.do_action_with_exception(request)
    return json.loads(str(response, encoding='utf-8'))


def del_domain_record_by_id(client: AcsClient, record_id: str) -> dict:
    request = DeleteDomainRecordRequest()
    request.set_accept_format('json')
    request.set_RecordId(record_id)
    response = client.do_action_with_exception(request)
    return json.loads(str(response, encoding='utf-8'))


class Handler(_DNSProviderBase):

    def __init__(self, access_key: str, secret: str, *args, **kwargs) -> None:
        self._aliyun_record_id: List[str] = []
        self.access_key = access_key
        self.secret =  secret

    def add_txt_record(self, 
                       identifier: str, 
                       value: str, 
                       *args, 
                       **kwargs) -> None:
        client = create_client(self.access_key, self.secret)
        if "*" in identifier:
            # if wildcard domain, remove "*"
            identifier = identifier.split('*.', maxsplit=1)[1]
        # aliyun takes non-punycode domain, change punycode back to literal
        domain = bytes(identifier, encoding='utf-8').decode('idna')
        resp_dict = add_dns_txt_record(
            client=client, 
            domain=domain,
            rr='_acme-challenge',
            value=value
        )
        self._aliyun_record_id.append(resp_dict['RecordId'])
    
    def clear_dns_record(self, *args, **kwargs) -> None:
        if self._aliyun_record_id:
            client = create_client(self.access_key, self.secret)
            for id in self._aliyun_record_id:
                del_domain_record_by_id(client, id)
                info(f'aliyun dns record {id} cleared')
            # reset aliyun record id to prevent multiple clear
            self._aliyun_record_id.clear()
        else:
            info('no aliyun dns record cleared')


from pyacme.settings import LETSENCRYPT_STAGING
import pytest

from conftest import Path
from conftest import settings
from test_common import *


@pytest.fixture(scope='class')
def root_host_entry(request):
    marker = request.node.get_closest_marker('host_entry')
    if marker is None:
        return
    add_host_entry(*marker.args)


@pytest.fixture(scope='class')
def setup_pebble_docker(request, root_host_entry):
    # override start_pebble_docker form conftest.py
    marker = request.node.get_closest_marker('docker_type')
    if marker.args[0] == 'standalone':
        print('using standalone container setup')
        # override start_pebble_docker() from conftest.py;
        # only run the pebble container, without challtest
        container_name = time.strftime(settings.BAK_TIME_FMT, time.localtime())
        container_name = 'pebble_' + container_name
        run_pebble_standalone_container(container_name)
        print(f'running container {container_name}')
        yield
        stop_pebble_standalone_container(container_name)
    elif marker.args[0] is None:
        # do not run any container
        yield


@pytest.fixture(scope='function')
def aliyun_access_key() -> Dict[str, str]:
    try:
        with open('./.aliyun_dns_api.json') as f:
            return json.load(f)
    except FileNotFoundError:
        try:
            with open('test/.aliyun_dns_api.json') as f:
                return json.load(f)
        except FileNotFoundError:
            # for ci test
            return {
                'access_key': os.environ['ALIYUN_AK'],
                'secret': os.environ['ALIYUN_S'],
            }


# def get_aliyun_access_key(key_file: str) -> Dict[str, str]:
#     with open(key_file, 'r') as f:
#         return json.load(f)


def run_test_main(**main_param) -> None:
    run_arg = []
    param_dict = {
        'country_code': '-C',
        'csr_subjects': '--csr_subjects',
        'account_private_key': '--account_private_key',
        'not_before': '--not_before',
        'not_after': '--not_after',
        'working_directory': '-w',
        'mode': '-m',
        'dnsprovider': '--dnsprovider',
        'access_key': '-k',
        'secret': '-s',
        'dns_specifics': '--dns_specifics',
        'CA_entry': '--CA_entry',
        'poll_interval': '--poll_interval',
        'poll_retry_count': '--poll_retry_count',
        'csr_priv_key_type': '--csr_priv_key_type',
        'csr_priv_key_size': '--csr_priv_key_size',
        'chall_resp_server_port': '--chall_resp_server_port',
    }

    for d in main_param['domain']:
        run_arg += ['-d', d]
    del main_param['domain']

    for c in main_param['contact']:
        run_arg += ['-c', c]
    del main_param['contact']

    if ('no_ssl_verify' in main_param) and main_param['no_ssl_verify']:
        run_arg += ['--no_ssl_verify']
        del main_param['no_ssl_verify']

    if ('debug' in main_param) and main_param['debug']:
        run_arg += ['--debug']
        del main_param['debug']

    for k, v in main_param.items():
        run_arg += [param_dict[k], v]

    main_entry_point_test(run_arg)


def _common(params: dict, ca = 'pebble'):
    # assert subprocess_run_pyacme(**params).returncode == 0
    run_test_main(**params)
    if 'working_directory' in params:
        wd = Path(params['working_directory']).expanduser().absolute()
    else:
        wd = Path(settings.WD_DEFAULT).expanduser().absolute()
    wd = wd / '_'.join(params['domain'])    # domain must exist
    root_cert = 'pebble-root-cert.pem'
    if ca == 'pebble':
        download_root_cert(wd/settings.WD_CERT)
        root_cert = 'pebble-root-cert.pem'
    elif ca == 'staging':
        root_cert = 'fake_root.pem'
        download_root_cert(wd/settings.WD_CERT, STAGING_ROOT_CA, root_cert)
    verify_p = openssl_verify(
        cert_path=wd/settings.WD_CERT/settings.CERT_NAME,
        chain_path=wd/settings.WD_CERT/settings.CERT_CHAIN,
        root_cert_path=wd/settings.WD_CERT,
        root_cert_name=root_cert
    )
    assert verify_p.returncode == 0


_DOMAIN = ['test-integration.local']
_MULTI_DOMAIN = ['a.test-integration.local', 'b.test-integration.local']
_HTTP_MODE_COMMON_PARAM_PORTION = dict(
    contact=TEST_CONTACTS,
    country_code='UN',
    CA_entry=PEBBLE_TEST,
    mode='http',
    no_ssl_verify=True,
    chall_resp_server_port=PY_HTTPSERVER_PORT,
)

http_mode_params = [
    pytest.param(
        dict(domain=_DOMAIN, **_HTTP_MODE_COMMON_PARAM_PORTION),
        id='http_mode_single_domain'
    ),
    pytest.param(
        dict(domain=_MULTI_DOMAIN, **_HTTP_MODE_COMMON_PARAM_PORTION),
        id='http_mode_multi_domain'
    ),
    pytest.param(
        dict(
            domain=_MULTI_DOMAIN, 
            working_directory='~/.pyacme/new', 
            **_HTTP_MODE_COMMON_PARAM_PORTION
        ),
        id='http_mode_multi_domain_new_wd'
    ),
    pytest.param(
        dict(
            domain=_DOMAIN, 
            **_HTTP_MODE_COMMON_PARAM_PORTION,
            csr_priv_key_type='es256'
        ),
        id='http_mode_single_domain_es256'
    ),
    pytest.param(
        dict(
            domain=_MULTI_DOMAIN, 
            **_HTTP_MODE_COMMON_PARAM_PORTION,
            csr_priv_key_type='es256'
        ),
        id='http_mode_multi_domain_es256'
    ),
]


@pytest.mark.httptest
@pytest.mark.docker_type('standalone')
@pytest.mark.host_entry(_DOMAIN+_MULTI_DOMAIN, '127.0.0.1')
@pytest.mark.usefixtures('setup_pebble_docker')
class TestHttpMode:
    @pytest.mark.parametrize('params', http_mode_params)
    def test_http_mode(self, params):
        _common(params)
        # wait for a while if this is to be tested with test_pebble.py
        # time.sleep(5)


_STAGING_DOMAIN = ['test-staging.xn--ihqz7no5gol3b.icu']
_STAGING_WILDCARD_DOMAIN = ['*.xn--ihqz7no5gol3b.icu']
_STAGING_MULTI_DOMAIN = [
    'test-staging-1.xn--ihqz7no5gol3b.icu',
    'test-staging-2.xn--ihqz7no5gol3b.icu',
]
_DNS_MODE_COMMON_PARAM_PORTION = dict(
    contact=TEST_CONTACTS,
    country_code='UN',
    CA_entry=LETSENCRYPT_STAGING,
    mode='dns',
)
_DNS_MODE_PEBBLE_PARAM: dict = _DNS_MODE_COMMON_PARAM_PORTION.copy()
_DNS_MODE_PEBBLE_PARAM.update(
    dict(
        CA_entry=PEBBLE_TEST,
        no_ssl_verify=True,
    )
)

dns_mode_params = [
    pytest.param(
        dict(domain=_STAGING_DOMAIN, **_DNS_MODE_COMMON_PARAM_PORTION),
        id='dns_mode_single_domain'
    ),
    pytest.param(
        dict(domain=_STAGING_MULTI_DOMAIN, **_DNS_MODE_COMMON_PARAM_PORTION),
        id='dns_mode_multi_domain'
    ),
    pytest.param(
        dict(domain=_STAGING_WILDCARD_DOMAIN, **_DNS_MODE_COMMON_PARAM_PORTION),
        id='dns_mode_wildcard_domain'
    ),
]

@pytest.mark.dnstest
@pytest.mark.docker_type(None)
@pytest.mark.usefixtures('setup_pebble_docker')
class TestDNSMode:

    @pytest.mark.parametrize('params', dns_mode_params)
    def test_dns_mode(self, params, aliyun_access_key: Dict[str, str]):
        key_dict = dict(
            access_key=aliyun_access_key['access_key'],
            secret=aliyun_access_key['secret']
        )
        params = dict(**params, **key_dict)
        _common(params, ca='staging')
        time.sleep(5)


@pytest.mark.dnstest_pebble
@pytest.mark.docker_type('standalone')
@pytest.mark.usefixtures('setup_pebble_docker')
class TestDNSModePebble:

    @pytest.mark.parametrize('key_type', ['rsa', 'es256'])
    def test_dns_mode_pebble(self, key_type, aliyun_access_key: Dict[str, str]):
        key_dict = dict(
            access_key=aliyun_access_key['access_key'],
            secret=aliyun_access_key['secret']
        )
        params = dict(**_DNS_MODE_PEBBLE_PARAM, **key_dict)
        params.update(dict(domain=_STAGING_DOMAIN, csr_priv_key_type=key_type))
        _common(params, ca='pebble')